;--------------------------------------
; Timer routines, 16-bit small model
;
; 1 tick of the 8253 timer = 0.0000008381 seconds
;                          = 838.1 nano seconds
;                          = 0.8381 x 10^(-6)
;                          = 0.8381 sec
;
; maximum count: 65535 x 8.381 x 10(-7) = 0.0549 seconds (about 55 msec)
;
; For accurate measurements interrupts must be disabled during
; the timing period. This also means that the normal BIOS/DOS
; clock time may be "suspended" for a short amount of time.
;
; Proper one-shot use scenerio:
;
; call timer_init
; ....
; cli                ; disable interrupts
; call timer_on
; call testcode      ; your code to test
; call timer_off
; sti                ; enable interrupts
; ...
; call timer_show
;
; Proper repeated event scenerio:
;
;   call timer_init
; foo:
;   ....
;   cli                ; disable interrupts
;   call timer_on
;   call testcode      ; your code to test
;   call timer_off
;   sti                ; enable interrupts
;   ...
;   loop foo
;   call timer_show_average
;
;--------------------------------------

include tm_model.inc

pic_read      equ 00001010b ; 8259 PIC read command
pic_ocw3      equ 20h       ; 8259 PIC operation command word 3
pic_int_req   equ 20h       ; 8259 PIC interrupt request register
timer_mode_2  equ 00110100b ; 8253 timer mode 2 command byte
timer_0_count equ 40h       ; 8253 timer 0 count register
timer_mode    equ 43h       ; 8253 mode register

delay macro
 ; this macro delays long enough to allow
 ; repeated I/O accesses to the same device
 ; without errors. This may not be needed on
 ; some machines and may need to be longer on others.
 jmp $+2
 jmp $+2
 jmp $+2
endm

delay_long macro
 ; This macro delays long enough to allow an interrupt handler to complete.
 delay
 delay
 delay
 delay
endm

load_ds macro reg       ; Load DS segment for compact, large and huge models
 ife @datasize
 else
  mov reg, @data
  mov ds, reg
 endif
endm

save_ds macro           ; Save DS segment for compact, large and huge models
 ife @datasize
 else
  push ds
 endif
endm

restore_ds macro        ; Restore DS segment for compact, large and huge models
 ife @datasize
 else
  pop  ds
 endif
endm

.data

public timed_count

timed_count        dw 0                ; counts for one timed event
total_count        dd 0                ; accumulated counts for many events
timer_status_array dw 0                ; number of events
                   dw 0                ; number of overflows
                   dw 0                ; number of underflows
                   dw 0                ; number of timer overhead errors
timer_error_flag   db 0                ; error code for last event
timer_result_fmt   db 0                ; 0 = ticks, 1 = microsec

out_string         db 5 dup(0)         ; space for 5 digit number
out_type           db 6 dup(' ')       ; space for ticks or sec string
                   db 13, 10           ; CR/LF
out_len equ $ - out_string

sec_str            db ' sec '
tick_str           db ' ticks'

% .code @codename

public timer_init
timer_init proc
;--------------------------------------
; Init timer variables
;
; Call any time before starting one or more timing sequences.
; Optional for timing a single event in a program.
;
; inputs: none
;
; outputs: none
;--------------------------------------

           push   ax

           save_ds
           load_ds ax

; zero cumulative count

           mov    word ptr total_count, 0
           mov    word ptr total_count+2, 0

; zero event count, error counts

           mov    timer_status_array, 0
           mov    timer_status_array+2, 0
           mov    timer_status_array+4, 0
           mov    timer_status_array+6, 0

           restore_ds
           pop    ax
           ret

timer_init endp


public timer_on
timer_on proc
;--------------------------------------
; Start the timer
;
; inputs: none
;
; outputs: none
;--------------------------------------

           push   ax
           save_ds

           load_ds ax

; clear error flag

           mov    timer_error_flag, 0

; turn on interrupts

           sti

; start timer

           call   timer_startup

; delay for a timer interrupt to complete

           delay_long

; disable interrupts

           cli

; restart timer again

           call   timer_startup

           restore_ds
           pop    ax
           ret

timer_on endp

public timer_off
timer_off proc
;--------------------------------------
; Stop the timer, save counts, check for overflow
;
; outputs: AX status
;             0   success
;             1   timer overflow (event too long)
;             2   timer underflow (event too short)
;             3   overhead too large
;--------------------------------------

           push   bx
           call   timer_stop           ; stop the timer, returns counts

           save_ds

; check for an un-serviced timer interrupt (indicates timer overflow)

           push   ax

           load_ds ax

           mov    al, pic_read         ; setup for PIC read
           out    pic_ocw3, al
           delay
           in     al, pic_int_req      ; read interrupt request register
           and    al, 1                ; keep only IRQ0 (timer interrupt)
           pop    ax
           jnz    tmr_err1

; determine true ticks by subtracting overhead ticks

           call   timer_overhead
           sub    bx, ax
           jb     tmr_err3             ; check for overhead calculation error
           cmp    bx, ax
           jbe    tmr_err2             ; check for timed event too small

; save the count

           mov    timed_count, bx               ; current counts
           add    word ptr total_count, bx      ; total cumulative counts
           adc    word ptr total_count+2, 0
           inc    timer_status_array            ; inc number of events
           mov    al, 0
tmr_ret:
           restore_ds
           pop    bx
           ret

; Error due to very large timed event
tmr_err1:
           mov    al, 1                ; error due to timer overflow
tmr_err:
           mov    timer_error_flag, al
           xor    bx, bx
           mov    bl, al
           add    bx, bx
           inc    timer_status_array[bx]
           jmp    tmr_ret

; Error due to very small timed event
tmr_err2:
           mov    al, 2
           jmp    tmr_err

; Error due to overhead > timed event. This shouldn't happen but randomly
; could if overhead is high and event being timed is very small

tmr_err3:
           mov    al, 3
           jmp    tmr_err

timer_off endp

public timer_startup
timer_startup proc
;--------------------------------------
; Start timer 0 in mode 2 with count of zero
; (internal routine)
;
; modifies AX
;--------------------------------------

           mov    al, timer_mode_2
           out    timer_mode, al       ; set timer mode
           delay

           xor    al, al
           out    timer_0_count, al    ; low byte
           delay
           out    timer_0_count, al    ; high byte

           ret

timer_startup endp

public timer_stop
timer_stop proc
;--------------------------------------
; Stop timer, return counts
; (internal routine)
;
; outputs: BX counts
;--------------------------------------

           push   ax
           mov    al, 0
           out    timer_mode, al       ; latch timer
           delay

           in     al, timer_0_count    ; get low byte
           delay
           mov    bl, al
           in     al, timer_0_count    ; get high byte
           mov    bh, al
           neg    bx                   ; (negate because timer is countdown)
           pop    ax
           ret

timer_stop endp

public timer_overhead
timer_overhead proc
;--------------------------------------
; Calculate the overhead for the timer
; (internal routine)
;
; inputs: none
;
; outputs: AX overhead ticks per start/stop of timer
;
;--------------------------------------

           push   bx
           push   cx
           mov    cx, 8                ; 8 loops
           xor    ax, ax               ; zero total counts
           cli                         ; disable interrupts
null_loop:
           push   ax                   ; pushes and pops in this loop
           push   ds                   ; are designed to mimic the
           call   timer_startup        ; pushes and pops in the real timer
           pop    ds                   ; code
           pop    ax
           call   timer_stop
           add    ax, bx
           loop   null_loop

           sti                         ; enable interrupts

           add    ax, 4                ; add 4 to rounding
           mov    cl, 3                ; divide by 8
           shr    ax, cl
           pop    cx
           pop    bx
           ret

timer_overhead endp


public timer_show
timer_show proc
;--------------------------------------
; Show timer results on screen in default format
; (show results of a single event)
;
; inputs: none
;
;--------------------------------------

           push   ax
           push   cx
           save_ds

           load_ds ax

           mov    cl, timer_result_fmt
           mov    ax, timed_count
           call   timer_write_default

           restore_ds
           pop    cx
           pop    ax
           ret

timer_show endp

public timer_write_default
timer_write_default proc
;--------------------------------------
; Show timer results on screen
;
; inputs: AX ticks
;         CL format
;
;--------------------------------------

           push   bx
           push   di
           push   si
           push   es

           load_ds bx
ife @datasize
else
           mov    es, bx
endif

; setup out_type string to 'ticks' or 'sec', if required convert to sec

           lea    di, out_type
           lea    si, tick_str
           test   cl, cl
           jnz    twd_1
           call   timer_ticks_to_microsec
           lea    si, sec_str
twd_1:
           mov    cx, 3                ; 6 bytes in out_type strings
           rep    movsw

           mov    bx, 1                ; stdout
           call   timer_write

           pop    es
           pop    si
           pop    di
           pop    bx
           ret

timer_write_default endp

public timer_show_ticks
timer_show_ticks proc
;--------------------------------------
; Show timer results on screen in ticks format
;
; inputs: none
;
;--------------------------------------

           push   cx
           save_ds
           load_ds ax
           mov    ax, timed_count
           mov    cl, 0                ; ticks format
           call   timer_write_default
           restore_ds
           pop    cx
           ret

timer_show_ticks endp

public timer_show_microseconds
timer_show_microseconds proc
;--------------------------------------
; Show timer results on screen in ticks format
;
; inputs: none
;
;--------------------------------------

           push   cx
           save_ds
           load_ds ax
           mov    ax, timed_count
           mov    cl, 1                ; microsecond format
           call   timer_write_default
           restore_ds
           pop    cx
           ret

timer_show_microseconds endp


public timer_show_average
timer_show_average proc
;--------------------------------------
; Show average timer ticks or microseconds on screen
; (show results of many timed events in default format)
;
; inputs: none
;
; outputs: CF set if division error
;          AX error code
;             1  no events (div by 0)
;             2  sum of times too large
;
;--------------------------------------

           push   cx
           push   dx
           save_ds

           load_ds ax

           mov    ax, word ptr total_count
           mov    dx, word ptr total_count+2

           mov    cx, timer_status_array       ; get count of events
           test   cx, cx
           jz     tsa_err1                     ; error if no events
           cmp    dx, cx
           jae    tsa_err2                     ; error if possible divide ovfl

           div    cx
           mov    cl, timer_result_fmt

           call   timer_write_default
           xor    ax, ax
tsa_ret:
           restore_ds
           pop    dx
           pop    cx
           ret

tsa_err1:
           mov    ax, 1
           jmp    short tsa_err
tsa_err2:
           mov    ax, 2
tsa_err:
           stc
           jmp    tsa_ret

timer_show_average endp

public timer_write
timer_write proc
;--------------------------------------
; Write timer result to (file) handle
;
; inputs: BX file handle for output (i.e. 1 = stdout)
;         AX ticks
;
;--------------------------------------

           push   ax
           push   bx
           push   cx
           push   dx
           push   si
           save_ds

           load_ds cx

           lea    si, out_string
           call   timer_ticks_to_ascii

           mov    dx, si               ; write to specified handle
           mov    cx, out_len
           mov    ah, 40h
           int    21h

           restore_ds
           pop    si
           pop    dx
           pop    cx
           pop    bx
           pop    ax
           ret

timer_write endp

public timer_ticks_to_ascii
timer_ticks_to_ascii proc
;--------------------------------------
; Convert timer ticks or microseconds to right justified ASCII string
;
; inputs: DS:SI ptr to 5 char buffer for ASCII string
;         AX    timer ticks
;
;--------------------------------------

           push   ax
           push   bx
           push   cx
           push   dx
           push   si

           add    si, 4                ; start by pointing to units

           mov    bx, 10               ; base 10 conversion
           mov    cx, 5
ttta_1:
           xor    dx, dx
           div    bx
           add    dl, '0'              ; convert remainder to ASCII
           mov    [si], dl             ; store number
           dec    si                   ; move to next higher digit
           or     ax, ax
           loopne ttta_1

           jcxz   ttta_ret
ttta_2:
           mov    byte ptr [si], ' '
           dec    si
           loop   ttta_2

ttta_ret:
           pop    si
           pop    dx
           pop    cx
           pop    bx
           pop    ax
           ret

timer_ticks_to_ascii endp


public timer_ticks_to_microsec
timer_ticks_to_microsec proc
;--------------------------------------
; Convert timer ticks to micro seconds
;
; inputs: AX timer ticks
;
; outputs: AX microseconds
;
;--------------------------------------

           push   bx
           push   dx

           mov    dx, 8381
           mul    dx
           mov    bx, 10000
           div    bx

           pop    dx
           pop    bx
           ret

timer_ticks_to_microsec endp

public timer_set_format
timer_set_format proc
;--------------------------------------
; Set the default timer result format
;
; inputs: AL format
;            0  clock tick format
;            1  microsecond format
;--------------------------------------

           push   bx
           save_ds

           load_ds bx

           mov    timer_result_fmt, al

           restore_ds
           pop    bx
           ret

timer_set_format endp

end
