;--------------------------------------
; Timer routines, 32-bit flat model
;
; See timer16.asm for documentation
; This 32-bit code is run in protected-mode (ring 3) so timings are not as
; accurate as 16-bit read-mode code because the protected-mode code does
; not disable interrupts.
;--------------------------------------

.386

pic_read      equ 00001010b ; 8259 PIC read command
pic_ocw3      equ 20h       ; 8259 PIC operation command word 3
pic_int_req   equ 20h       ; 8259 PIC interrupt request register
timer_mode_2  equ 00110100b ; 8253 timer mode 2 command byte
timer_0_count equ 40h       ; 8253 timer 0 count register
timer_mode    equ 43h       ; 8253 mode register

delay macro
 ; this macro delays long enough to allow
 ; repeated I/O accesses to the same device
 ; without errors. This may not be needed on
 ; some machines and may need to be longer on others.
 jmp $+2
 jmp $+2
 jmp $+2
endm

delay_long macro
 ; This macro delays long enough to allow an interrupt handler to complete.
 delay
 delay
 delay
 delay
endm

load_ds macro reg
endm

save_ds macro
endm

restore_ds macro
endm

msti macro
endm

mcli macro
endm

_data segment public use16 'data'
assume DS:_data

public timed_count

timed_count        dw 0                ; counts for one timed event
total_count        dd 0                ; accumulated counts for many events
timer_status_array dw 0                ; number of events
                   dw 0                ; number of overflows
                   dw 0                ; number of underflows
                   dw 0                ; number of timer overhead errors
timer_error_flag   db 0                ; error code for last event
timer_result_fmt   db 0                ; 0 = ticks, 1 = microsec

out_string         db 5 dup(0)         ; space for 5 digit number
out_type           db 6 dup(' ')       ; space for ticks or sec string
                   db 13, 10           ; CR/LF
out_len equ $ - out_string

sec_str            db ' sec '
tick_str           db ' ticks'

_data ends

code32  segment public para use32 'code'
assume CS:code32, DS:_data

public timer_init
timer_init proc
;--------------------------------------
; Init timer variables
;
; (Call any time before starting one or more timing sequences)
;
; inputs: none
;
; outputs: none
;--------------------------------------

           push   eax

           save_ds
           load_ds ax

; zero cumulative count

           mov    word ptr total_count, 0
           mov    word ptr total_count+2, 0

; zero event count, error counts

           mov    timer_status_array, 0
           mov    timer_status_array+2, 0
           mov    timer_status_array+4, 0
           mov    timer_status_array+6, 0

           restore_ds
           pop    eax
           ret

timer_init endp


public timer_on
timer_on proc
;--------------------------------------
; Start the timer
;
; inputs: none
;
; outputs: none
;--------------------------------------

           push   eax
           save_ds

           load_ds ax

; turn on interrupts

           msti

; start timer

           call   timer_startup

; delay for a timer interrupt to complete

           delay_long

; disable interrupts

           mcli

; restart timer again

           call   timer_startup

           restore_ds
           pop    eax
           ret

timer_on endp

public timer_off
timer_off proc
;--------------------------------------
; Stop the timer, save counts, check for overflow
;
; outputs: AX status
;             0   success
;             1   timer overflow (event too long)
;             2   timer underflow (event too short)
;             3   overhead too large
;--------------------------------------

           push   ebx
           call   timer_stop           ; stop the timer, returns counts

           save_ds

; check for an un-serviced timer interrupt (indicates timer overflow)

           push   ax

           load_ds ax

           mov    al, pic_read         ; setup for PIC read
           out    pic_ocw3, al
           delay
           in     al, pic_int_req      ; read interrupt request register
           and    al, 1                ; keep only IRQ0 (timer interrupt)
           pop    ax
           jnz    tmr_err1

; determine true ticks by subtracting overhead ticks

           call   timer_overhead
           sub    bx, ax
           jb     tmr_err3             ; check for overhead calculation error
           cmp    bx, ax
           jbe    tmr_err2             ; check for timed event too small

; save the count

           mov    timed_count, bx               ; current counts
           add    word ptr total_count, bx      ; total cumulative counts
           adc    word ptr total_count+2, 0
           inc    timer_status_array            ; inc number of events
           mov    al, 0
tmr_ret:
           restore_ds
           pop    ebx
           ret

; Error due to very large timed event
tmr_err1:
           mov    al, 1                ; error due to timer overflow
tmr_err:
           mov    timer_error_flag, al
           xor    ebx, ebx
           mov    bl, al
           add    ebx, ebx
           inc    timer_status_array[ebx]
           jmp    tmr_ret

; Error due to very small timed event
tmr_err2:
           mov    al, 2
           jmp    tmr_err

; Error due to overhead > timed event. This shouldn't happen but could
; if overhead is high and event being timed is very small

tmr_err3:
           mov    al, 3
           jmp    tmr_err

timer_off endp

public timer_startup
timer_startup proc
;--------------------------------------
; Start timer 0 in mode 2 with count of zero
; (internal routine)
;
; modifies AX
;--------------------------------------

           mov    al, timer_mode_2
           out    timer_mode, al       ; set timer mode
           delay

           xor    al, al
           out    timer_0_count, al    ; low byte
           delay
           out    timer_0_count, al    ; high byte

           ret

timer_startup endp

public timer_stop
timer_stop proc
;--------------------------------------
; Stop timer, return counts
; (internal routine)
;
; outputs: BX counts
;--------------------------------------

           push   eax
           mov    al, 0
           out    timer_mode, al       ; latch timer
           delay

           in     al, timer_0_count    ; get low byte
           delay
           mov    bl, al
           in     al, timer_0_count    ; get high byte
           mov    bh, al
           neg    bx                   ; (negate because timer is countdown)
           pop    eax
           ret

timer_stop endp

public timer_overhead
timer_overhead proc
;--------------------------------------
; Calculate the overhead for the timer
; (internal routine)
;
; inputs: none
;
; outputs: AX overhead ticks per start/stop of timer
;
;--------------------------------------

           push   ebx
           push   ecx
           mov    ecx, 8               ; 8 loops
           xor    ax, ax               ; zero total counts
           mcli                        ; disable interrupts
null_loop:
           push   ax                   ; pushes and pops in this loop
           push   ds                   ; are designed to mimic the
           call   timer_startup        ; pushes and pops in the real timer
           pop    ds                   ; code
           pop    ax
           call   timer_stop
           add    ax, bx
           loop   null_loop

           msti                        ; enable interrupts

           add    ax, 4                ; add 4 to round
           shr    ax, 3
           pop    ecx
           pop    ebx
           ret

timer_overhead endp


public timer_show
timer_show proc
;--------------------------------------
; Show timer results on screen in default format
; (show results of a single event)
;
; inputs: none
;
;--------------------------------------

           push   eax
           push   ecx
           save_ds

           load_ds ax

           mov    cl, timer_result_fmt
           mov    ax, timed_count
           call   timer_write_default

           restore_ds
           pop    ecx
           pop    eax
           ret

timer_show endp

public timer_write_default
timer_write_default proc
;--------------------------------------
; Show timer results on screen
;
; inputs: AX ticks
;         CL format
;
;--------------------------------------

           push   ebx
           push   edi
           push   esi
           push   es

           load_ds bx

           push   ds
           pop    es

; setup out_type string to 'ticks' or 'sec', if required convert to sec

           lea    edi, out_type
           lea    esi, tick_str
           test   cl, cl
           jnz    twd_1
           call   timer_ticks_to_microsec
           lea    esi, sec_str
twd_1:
           mov    ecx, 3               ; 6 bytes in out_type strings
           rep    movsw

           mov    bx, 1                ; stdout
           call   timer_write

           pop    es
           pop    esi
           pop    edi
           pop    ebx
           ret

timer_write_default endp

public timer_show_ticks
timer_show_ticks proc
;--------------------------------------
; Show timer results on screen in ticks format
;
; inputs: none
;
;--------------------------------------

           push   ecx
           save_ds
           load_ds ax
           mov    ax, timed_count
           mov    cl, 0                ; ticks format
           call   timer_write_default
           restore_ds
           pop    ecx
           ret

timer_show_ticks endp

public timer_show_microseconds
timer_show_microseconds proc
;--------------------------------------
; Show timer results on screen in ticks format
;
; inputs: none
;
;--------------------------------------

           push   ecx
           save_ds
           load_ds ax
           mov    ax, timed_count
           mov    cl, 1                ; microsecond format
           call   timer_write_default
           restore_ds
           pop    ecx
           ret

timer_show_microseconds endp


public timer_show_average
timer_show_average proc
;--------------------------------------
; Show average timer ticks or microseconds on screen
; (show results of many timed events in default format)
;
; inputs: none
;
; outputs: CF set if division error
;          AX error code
;             1  no events (div by 0)
;             2  sum of times too large
;
;--------------------------------------

           push   ecx
           push   edx
           save_ds

           load_ds ax

           mov    eax, dword ptr total_count

           xor    ecx, ecx
           xor    edx, edx
           mov    cx, timer_status_array       ; get count of events
           test   ecx, ecx
           jz     tsa_err1                     ; error if no events

           div    ecx
           mov    cl, timer_result_fmt

           call   timer_write_default
           xor    eax, eax
tsa_ret:
           restore_ds
           pop    edx
           pop    ecx
           ret

tsa_err1:
           mov    ax, 1
           jmp    short tsa_err
tsa_err2:
           mov    ax, 2
tsa_err:
           stc
           jmp    tsa_ret

timer_show_average endp

public timer_write
timer_write proc
;--------------------------------------
; Write timer result to (file) handle
;
; inputs: BX file handle for output (i.e. 1 = stdout)
;         AX ticks
;
;--------------------------------------

           push   eax
           push   ebx
           push   ecx
           push   edx
           push   esi
           save_ds

           load_ds cx

           lea    esi, out_string
           call   timer_ticks_to_ascii

           mov    edx, esi               ; write to specified handle
           mov    ecx, out_len
           mov    ah, 40h
           int    21h

           restore_ds
           pop    esi
           pop    edx
           pop    ecx
           pop    ebx
           pop    eax
           ret

timer_write endp

public timer_ticks_to_ascii
timer_ticks_to_ascii proc
;--------------------------------------
; Convert timer ticks or microseconds to right justified ASCII string
;
; inputs: DS:ESI ptr to 5 char buffer for ASCII string
;         AX     timer ticks
;
;--------------------------------------

           push   eax
           push   ebx
           push   ecx
           push   edx
           push   esi

           add    esi, 4               ; start by pointing to units

           mov    ebx, 10              ; base 10 conversion
           mov    ecx, 5
ttta_1:
           xor    dx, dx
           div    bx
           add    dl, '0'              ; convert remainder to ASCII
           mov    [esi], dl            ; store number
           dec    esi                  ; move to next higher digit
           or     ax, ax
           loopne ttta_1

           jcxz   ttta_ret
ttta_2:
           mov    byte ptr [esi], ' '
           dec    esi
           loop   ttta_2

ttta_ret:
           pop    esi
           pop    edx
           pop    ecx
           pop    ebx
           pop    eax
           ret

timer_ticks_to_ascii endp


public timer_ticks_to_microsec
timer_ticks_to_microsec proc
;--------------------------------------
; Convert timer ticks to micro seconds
;
; inputs: AX timer ticks
;
; outputs: AX microseconds
;
;--------------------------------------

           push   ebx
           push   edx

           mov    dx, 8381
           mul    dx
           mov    bx, 10000
           div    bx

           pop    edx
           pop    ebx
           ret

timer_ticks_to_microsec endp

public timer_set_format
timer_set_format proc
;--------------------------------------
; Set the default timer result format
;
; inputs: AL format
;            0  clock tick format
;            1  microsecond format
;--------------------------------------

           push   ebx
           save_ds

           load_ds bx

           mov    timer_result_fmt, al

           restore_ds
           pop    ebx
           ret

timer_set_format endp

code32 ends

end
