// This file provide you the routines
// for conversion color from various
// color models to RGB and back.
// It manages the following conversions:
//	CIE2RGB <-> RGB2CIE
//	HLS2RGB <-> RGB2HLS
//	HSV2RGB <-> RGB2HSV
//	YIQ2RGB <-> RGB2YIQ
//	RGB2CMY
//	RGB2CMYK
// (c) This program is written by Kiselev J. 1994.

#include <math.h>
#include "colmodel.h"
#define UNDEFINED -400

float value (float a,float b,float hue) {
	if (hue > 360) hue = hue - 360;
	else if (hue < 0) hue = hue + 360;
	if (hue < 60) return(a + (b - a) * hue/60);
	else if (hue < 180) return(b);
	else if (hue < 240) return(a + (b - a) * (240 - hue)/60);
	else return(a);
}

void HLS2RGB(float H,float L,float S,float &R,float &G,float &B) {
	float m,n;
	if (L <= 0.5) n = L * (1 + S);
	else n = L + S - L * S;
	m = 2 * L - n;
	if ((S == 0) && (H == -400)) {R = L; G = L; B = L;}
	else {
		R = value(m,n,H+120);
		G = value(m,n,H);
		B = value(m,n,H-120);
	}
}

void RGB2HLS(float R,float G,float B,float &H,float &L,float &S) {
	float min,max,delta;

	max = max3(R,G,B);
	min = min3(R,G,B);
	L = (max + min) / 2;
	if (max == min) {S = 0; H = UNDEFINED;}
	else	{
		delta = max - min;
		if (L <= 0.5) S = delta / (max + min);
		else S = delta / (2 - max - min);
		if (R == max) H = (G - B) / delta;
		else
		if (G == max) H = 2 + (B - R) / delta;
		else
		if (B == max) H = 4 + (R - G) / delta;
		H *= 60;
		if (H < 0)  H += 360;
	}
}

void HSV2RGB(float H,float S,float V,float &R,float &G,float &B) {
	float f,j,k,l;
	long i;
	if (S == 0) if (H == -400) {R = V; G = V; B = V; return;} else return;
	if (H < 0) H += 360;
	if (H == 360) H = 0; else H = H/60;
	i = floor(H);
	f = H - i;
	j = V * (1 - S);
	k = V * (1 - S * f);
	l = V * (1 - S * (1 - f));

	switch (i) {
		case 0 : R = V; G = l; B = j; break;
		case 1 : R = k; G = V; B = j; break;
		case 2 : R = j; G = V; B = l; break;
		case 3 : R = j; G = k; B = V; break;
		case 4 : R = l; G = j; B = V; break;
		case 5 : R = V; G = j; B = k; break;
	}
}

void RGB2HSV(float R,float G,float B,float &H,float &S,float &V) {
	float Cr,Cg,Cb,min,d;
	V = max3(R,G,B);
	min = min3(R,G,B);
	if (V != 0) S = (V - min) / V; else S = 0;
	if (S == 0) H = UNDEFINED;
	else
	{
		d = V - min;
		Cr = (V - R)/d;
		Cg = (V - G)/d;
		Cb = (V - B)/d;
		if (R == V) H = Cb - Cg;
		else
		if (G == V) H = 2 + Cr - Cb;
		else
		if (B == V) H = 4 + Cg - Cr;
		H *= 60;
		if (H < 0) H = H + 360;
	}
}

void YIQ2RGB(float Y,float I,float Q,float &R,float &G,float &B) {
	R = Y + 0.956*I + 0.623*Q;
	G = Y - 0.272*I - 0.648*Q;
	B = Y - 1.105*I + 1.705*Q;
}

void RGB2YIQ(float R,float G,float B,float &Y,float &I,float &Q) {
Y = 0.299*R + 0.587*G + 0.114*B;
I = 0.596*R - 0.274*G - 0.322*B;
Q = 0.211*R - 0.522*G + 0.311*B;
}

void CIE2RGB(float x,float y,float z,float &R,float &G,float &B) {
	float x1,z1;
	x1 = x*z/y;
	z1 = (1 - x - y)*z/y;
	R = 2.739*x1 - 1.145*z - 0.424*z1;
	G = - 1.119*x1 + 2.029*z +0.033*z1;
	B = 0.138*x1 - 0.333*z + 1.105*z1;
}

void RGB2CIE(float R,float G,float B,float &x,float &y,float &z) {
	float x1,z1,sum;
	x1 = 0.478*R + 0.299*G + 0.175*B;
	z = 0.263*R + 0.655*G + 0.081*B;
	z1 = 0.020*R + 0.160*G + 0.908*B;
	sum = x1 + z + z1;
	x = x1/sum;
	y = z/sum;
}

void RGB2CMY(float R,float G,float B,float &C,float &M,float &Y) {
	C = 1 - R;
	M = 1 - G;
	Y = 1 - B;
}

void RGB2CMYK(float R,float G,float B,float &C,float &M,float &Y,float &K) {
	RGB2CMY(R,G,B,C,M,Y);
	K = max3(C,M,Y);
	C = C - K;
	M = M - K;
	Y = Y - K;
}

void mixcolors(float x1,float y1,float z1,float x2,float y2,float z2,
		float &x12,float &y12,float &z12) {
	float t1,t2;
	t1 = z1/y1; t2 = z2/y2;
	x12 = (x1*t1 + x2*t2)/(t1 + t2);
	y12 = (y1*t1 + y2*t2)/(t1 + t2);
	z12 = z1 + z2;
}
