// This program contains an implementation of
// various graphics file formats readers and savers
#pragma -ml

#include <mem.h>
#include <alloc.h>
#include <dir.h>
#include <string.h>
#include <stdlib.h>
#include "common.h"
#include "formats.h"
#include "formconv.h"
#include "graph.h"

int format_file_error = 0;
char * format_file_error_messages[] = {
	"Ok", //0
	"File not found", //1
	"Unknown file format", //2
	"Format header error", //3
	"Disk read / write error" }; //4

/*********************** TARGA ******************/
/*********************** TARGA ******************/
/*********************** TARGA ******************/

targa_file::targa_file(char *fname) : graph_file_reader(fname) {
	int i;
	WORD j;
	BYTE *pal;

	format_file_error = 1;
	if  (!file_ok) return;	// return if constructor can't open the file

	file_ok = FALSE;
	format_file_error = 3;

	i = fread(&header,sizeof(targa_header),1,the_file);	// get file header
	if (i <= 0 || header.image_type == 0) return; //return if file isn't targa
	fseek(the_file,header.ID_field_length,SEEK_CUR);

	height = header.height;
	width = header.width;

	top_to_bottom = header.top_to_bottom;
	left_to_right =  (header.right_to_left == 0);
	colors = 256;
	rle_comp = FALSE;
	line2standart = col2562byte; //just move
	switch(header.image_type) {
		case 1: source_type = COLOR256IMG; break;
		case 2: source_type = TRUECOLORIMG; break;
		case 3: source_type = GRAY256IMG; break;
		case 9: source_type = COLOR256IMG;
			rle_comp = TRUE;
			break;
		case 10:source_type = TRUECOLORIMG;
			rle_comp = TRUE;
			break;
		case 11:source_type = GRAY256IMG;
			rle_comp = TRUE;
			break;
		default : return;
		}
	if(source_type == GRAY256IMG) {
		for(i=0;i<256;i++) {
			palette[i].r = i;
			palette[i].g = i;
			palette[i].b = i;
		}
	}
	if(source_type == TRUECOLORIMG) {
		colors = (header.bitsperpixel <= 16) ? (WORD)1 << 15 :
							(DWORD)1L << 24;
	}
	if(header.color_map_type) {
		if (header.colmap_length > 256) return;
		pal = new BYTE [header.colmap_length*header.colmap_entry_size/8];
		fread(pal,header.colmap_length*header.colmap_entry_size/8,1,the_file);
		switch (header.colmap_entry_size) {
		case 24:
			memmove(&palette[header.first_colmap_entry],
						pal,header.colmap_length*3);
			break;
		case 16:
		case 15:
			for(i = header.first_colmap_entry;i<header.first_colmap_entry+
				header.colmap_length;i++) {
				j = (WORD)((pal[i*2+1] << 8) | (WORD)pal[i*2]);
				palette[i].r = (j >> 10) & 0x1F;
				palette[i].g = (j >> 5) & 0x1F;
				palette[i].b = j & 0x1F;
			} break;
		case 32:
			for(i=header.first_colmap_entry;
			i<header.first_colmap_entry+
			header.colmap_length;i++,j+=4) {
				palette[i].r = pal[j];
				palette[i].g = pal[j+1];
				palette[i].b = pal[j+2];
			}
		}
		delete pal;
	}

	data_begin = ftell(the_file);
	counter = 0;
	bytesperpixel = (header.bitsperpixel+1)/8;
	switch(bytesperpixel) {
	case 2:	source_ext_type = RGB15;
		line2standart = RGB15_2RGBlines;
		break;
	case 3:	source_ext_type = BGR24;
		line2standart = BGR2RGBlines;
		break;
	case 4:	source_ext_type = BGRA32;
		line2standart = BGRA32_2RGBlines;
		break;
	}
	bytesperline = bytesperpixel*header.width;
	source_line = new BYTE[bytesperline+128*bytesperpixel];
	if (source_type == TRUECOLORIMG)
		standart_line = new BYTE[width*3];
	else standart_line = new BYTE[width];
	file_ok = TRUE;
	format_file_error = 0;
	init_buffer( rle_comp );
	seek_start();
}

int targa_file::get_next_line() {
	BYTE a,buff[4];
	int i,j;
	if(line_n < 0 || line_n >= height) return(-1);
	if(rle_comp) {
		if(counter != 0) memmove(source_line,
				&source_line[bytesperline],counter);
		fresh_buffer();

		while(counter<bytesperline) {
			a = buffer[buff_offset++];
			if ((a & 0x80) != 0) {
				for (i = 0; i < bytesperpixel; i++) buff[i] = buffer[buff_offset++];
				for (i = (a & 0x7F) + 1; i > 0; i--)
					for(j = 0; j < bytesperpixel; j++)
						source_line[counter++] = buff[j];
			}
			else { //for(i=0;i<(a+1)*bytesperpixel;i++)
			      //	source_line[counter++] = buffer[buff_offset++];
			      i = (a+1)*bytesperpixel;
			      memcpy(&source_line[counter], &buffer[buff_offset],i);
			      counter += i; buff_offset += i;
                        }
		}
		counter -= bytesperline;
	}  // of rle compression
	else fread(source_line,bytesperline,1,the_file);

	if (top_to_bottom) line_n++; else line_n--;
	return(0);
}


targa_file::~targa_file() {
	if (source_line) delete source_line;
	if (standart_line) delete standart_line;
}

targa_file_saver::targa_file_saver(char *fname, image_type dest_type_,
			image_extention_type dest_ext_type_,
			int width_,int height_,BGRpalette* pal,BOOL compress,
			BOOL top_to_bottom_,BOOL left_to_right_) :
	graph_file_saver(fname,compress, dest_type_,dest_ext_type_,
					width_,height_,pal)    		{
	int i;

	rle_comp = FALSE;
	format_file_error = 1;
	if  (!file_ok) return;	// return if constructor can't open the file

	format_file_error = 3;
	file_ok = FALSE;
	top_to_bottom = top_to_bottom_;
	left_to_right = left_to_right_;
	rle_comp = compress;
	header.width = width;
	header.height = height;
	header.color_map_type = header.ID_field_length = 0;
	header.attrbits = header.interleave = header.x = header.y = 0;

	line2standart = col2562byte; //just move
	colors = 256;

	switch(dest_type) {
	case MONOIMG:    dest_type = GRAY256IMG;

	case GRAY256IMG: header.image_type = rle_comp ? 11 : 3;
			 header.bitsperpixel = 8;
			 dest_ext_type = NO_EXTENTION;
			 break;
	case COLOR16IMG: dest_type = COLOR256IMG;
	case COLOR256IMG:header.image_type = rle_comp ? 9 : 1;
			 header.color_map_type = 1;
			 header.bitsperpixel = 8;
			 break;
	case TRUECOLORIMG:header.image_type = rle_comp ? 10 : 2;
			  switch(dest_ext_type) {
			  case RGB15:	header.bitsperpixel = 16;
					header.attrbits = 1;
					line2standart = RGBlines2RGB15;
					colors = 1<<15;
					break;
			  case NO_EXTENTION:
			  case BGR24:	header.bitsperpixel = 24;
					line2standart = RGBlines2BGR;
					colors = 1L << 24;
					break;
			  case BGRA32:  header.bitsperpixel = 32;
					header.attrbits = 8;
					line2standart = RGBlines2BGRA32;
					colors = 1L << 24;
					break;
			  default :     dest_ext_type = BGR24;
					header.bitsperpixel = 24;
					line2standart = RGBlines2BGR;
					colors = 1L << 24;
					break;
			  }
			  break;
	default : return;
	}
	if(header.color_map_type == 1) {
		header.first_colmap_entry = 0;
		header.colmap_length = 256;	// save standart VGA palette only!!!
		header.colmap_entry_size = 24;
	}
	else header.first_colmap_entry = header.colmap_length =
					header.colmap_entry_size = 0;


	if (top_to_bottom)  header.top_to_bottom = 1;
	else header.top_to_bottom = 0;
	if (left_to_right)  header.right_to_left = 0;
	else header.right_to_left = 1;


	line_n = top_to_bottom ? 0 : height-1;
	count_row = 0;
	bytesperline = width*((header.bitsperpixel+1)/8);	// +1 for 15 bits = 2 bytes
	bytesperpixel = (header.bitsperpixel+1)/8;

	if(rle_comp) arr_ofs = new int [width];
	fwrite(&header,sizeof(targa_header),1,the_file);
	if(header.color_map_type == 1) fwrite(palette,sizeof(VGApalette),1,the_file);
	file_ok = TRUE;
	format_file_error = 0;
}

void targa_file_saver::update_palette(BGRpalette *pal) {
	palette = pal;
	if (header.color_map_type != 1) return;
	fseek(the_file,sizeof(targa_header),SEEK_SET);
	fwrite(pal,sizeof(BGRpalette),1,the_file);
	fseek(the_file,0,SEEK_END);
}

targa_file_saver::put_next_line(BYTE* line) {
	int i,j,k,k1,l,counter=0,offset,nextoffset,bytesperpixel1;

	BYTE *lp;
	BOOL r = FALSE; 	      // repeated pixels are scanned
	DWORD buff1,buff2;

	count_row++;
	if (top_to_bottom) line_n++; else line_n--;

	if(rle_comp) {
		// mark starts of repeated  and non - repeated pixel sequences's
		bytesperpixel1 = bytesperpixel - 1;
		lp = line;
		switch (bytesperpixel) {

		case 1:
			for(j = 0; j < width-1; j++)
			if (r ^ ((*lp++) == *lp)) {
				r = !r;
				if (r) arr_ofs[counter++] = j;   // start of repeated pixels
					 // else finish of repeated, next - start of non-repeated
				else   arr_ofs[counter++] = j+1;
				}
			break;

		case 2:
		case 3:
		case 4:
			// i - bytes index
			buff1 = *(lp++);
			for (i = bytesperpixel1; i > 0; i--) buff1 = (buff1 << 8) + *(lp++);
			// buff1 and buff2 - two adjacent pixels
			for(j = 0; j < width-1; j++) {   // j - pixels index
				buff2 = *(lp++);
				for (i = bytesperpixel1; i > 0; i--) buff2 = (buff2 << 8) + *(lp++);
				if (r ^ (buff1 == buff2)) {
					r = !r;
					if (r) arr_ofs[counter++] = j;   // start of repeated
					 // else finish of repeated, next - start of non-repeated
					else   arr_ofs[counter++] = j+1;
				}
				buff1 = buff2;
			}
			break;

		default: return(-1);
		}

		arr_ofs[counter] = width; // force finish

		// we are starting from non-repeated seq. (may be lenght of 0)
		// and scan arr_ofs
		r = FALSE;
		counter = offset = 0;
		while (offset < width) {
			nextoffset =  arr_ofs[counter++];
			l = nextoffset - offset;
			if (r) 	while (l > 0) { // repeated
				k = min(128,l);
				l -= k;
				buffer[buff_offset++] = (k - 1) | 0x80;
				memcpy(&buffer[buff_offset],&line[offset*bytesperpixel],bytesperpixel);
				buff_offset += bytesperpixel;
				}
			else while (l > 0) {   // non-repeated
				k = min(128,l);
				l -= k;
				buffer[buff_offset++] = k - 1;
				k1 = k*bytesperpixel;
				memcpy(&buffer[buff_offset],&line[offset*bytesperpixel],k1);
				offset += k;
				buff_offset += k1;
				}
			offset = nextoffset;
			r = !r;
			}

		if (buff_offset > (BUFF_SIZE - (bytesperline+bytesperline/2)) || count_row == height)
				flush_buffer();
	}
	else fwrite(line,bytesperline,1,the_file);
	return(0);
}

targa_file_saver::~targa_file_saver() {
	if(rle_comp) delete arr_ofs;
}


/***********************  BMP ******************/
/***********************  BMP ******************/
/***********************  BMP ******************/

bmp_file::bmp_file(char *fname) : graph_file_reader(fname) {
	BYTE bmp_col[4];
	long fsize,line_len;
	int i;

	format_file_error = 1;
	if  (!file_ok) return;

	init_buffer(FALSE);

	format_file_error = 3;
	file_ok = FALSE;
	i = fread(&header, sizeof(header), 1, the_file);
	if ((i<=0) || (header.bf_type != BMPSIGNATURE)) return;
	data_begin = header.bf_offbits;

	if (fread(&info, sizeof(info), 1, the_file) <= 0) return;
	if (info.bi_compression) return;

	fseek(the_file,info.bi_size- sizeof(info), SEEK_CUR);

	height = info.bi_height;
	width = info.bi_width;
	top_to_bottom  = FALSE;

	line2standart = col2562byte;
	switch(info.bi_bit_count) {
		case 1: source_type = MONOIMG;
			colors = 2;
			line2standart = mono2mono;
			break;
		case 4: source_type = COLOR16IMG;
			colors = 16;
			source_ext_type = COLOR16_2PIXEL;
			line2standart = col16A2col16B;
			break;
		case 8: source_type = COLOR256IMG; colors = 256; break;
		case 24: source_type = TRUECOLORIMG;
			colors = 1L<<24;
			source_ext_type = BGR24;
			line2standart = BGR2RGBlines;
		}
	if (source_type != TRUECOLORIMG) {
		for (i = 0; i < min(256,colors); i++) {
			if (fread(bmp_col,4,1,the_file) <= 0) return;
			memmove(&palette[i],bmp_col,3);
		}
	}
	if (info.bi_clr_used) colors = info.bi_clr_used;

	fseek(the_file,0,SEEK_END); fsize = ftell(the_file);
	bytesperline = (fsize - header.bf_offbits) / info.bi_height;
		//alas, it's the only way to know actual line size!!!

	seek_start();
	source_line = new BYTE[bytesperline];
	switch  (source_type) {
	case MONOIMG:
	case COLOR256IMG:
	case GRAY256IMG:
	case TRUECOLORIMG:
		standart_line = new BYTE[bytesperline];
		break;
	case COLOR16IMG:
		standart_line = new BYTE[2*bytesperline];
		break;
	}

	file_ok = TRUE;
	format_file_error = 0;

}


bmp_file::~bmp_file() {
	if (source_line) delete source_line;
	if (standart_line) delete standart_line;
}

int bmp_file::get_next_line() {
	if ((line_n-- < 0) ||
	     (fread(source_line, bytesperline, 1, the_file)) <= 0) return(-1);
	return(0);
}

void build_bmp_info(bmp_file_info_header &info,
	    image_type f_type,	long width,long height) {

	info.bi_size = sizeof(info);
	info.bi_width = width;
	info.bi_height = height;
	info.bi_planes = 1;
	info.bi_compression = 0;
	info.bi_xpels_per_meter = 90*(10000 / 254);
	info.bi_ypels_per_meter = info.bi_xpels_per_meter;

	info.bi_clr_used = 0;
	info.bi_clr_important = 0;
	info.bi_bit_count = get_bitsperpixel(f_type,NO_EXTENTION);

	info.bi_size_image = (info.bi_bit_count*width+7)/8*height;
}

void bmp_file_saver::init(bmp_file_info_header *info_) {
	bmp_file_header header;
	int i;

	line2standart = col2562byte;
	switch  (dest_type) {
	case GRAY256IMG : dest_ext_type = GRAY_PALETTE;
			break;
	case MONOIMG: line2standart = mono2mono;
			 break;
	case COLOR16IMG: dest_ext_type =   COLOR16_2PIXEL;
			 line2standart = col16B2col16A;
			 break;
	case TRUECOLORIMG :  dest_ext_type =   BGR24;
			     line2standart = RGBlines2BGR;
			     break;
	default: dest_ext_type =  NO_EXTENTION;
	}


	if (!file_ok) return;
	file_ok = FALSE;
	top_to_bottom = FALSE;
	line_n = height-1;

	info = *info_;
	bytesperline = (((width*info.bi_bit_count+7) / 8 + 3) /4)*4;
	colors = 1L << info.bi_bit_count;

	header.bf_type = BMPSIGNATURE;

	header.bf_offbits = sizeof(header) + sizeof(bmp_file_info_header);

	if (colors < 257) header.bf_offbits +=  colors * 4;

	header.bf_size = ((long)bytesperline*height+(long)header.bf_offbits);
	header.bf_reserved1 = header.bf_reserved2 = 0;


	if (fwrite(&header,sizeof(header),1,the_file) <= 0) return;
	if (fwrite(&info,sizeof(info),1,the_file) <= 0) return;

	file_ok = TRUE;
	format_file_error = 0;
	update_palette(palette);
}

bmp_file_saver::bmp_file_saver(char *fname,image_type dest_type_,
			int width_,int height_,BGRpalette * pal):
	graph_file_saver(fname, FALSE, dest_type_, NO_EXTENTION,
			width_, height_,pal) {
	bmp_file_info_header info;

	build_bmp_info(info,dest_type_,width_,height_);
	init(&info);
}

bmp_file_saver::bmp_file_saver(char *fname,image_type dest_type_,
			 BGRpalette *palette, bmp_file_info_header *info_):
	graph_file_saver(fname, FALSE, dest_type_, NO_EXTENTION,
			info_->bi_width, info_->bi_height,palette) {
	init(info_);
}

void bmp_file_saver::update_palette(BGRpalette *pal) {
	BYTE bmp_col[4];
	int i;

	if (dest_type == TRUECOLORIMG) return;
	fseek(the_file,sizeof(bmp_file_info_header)+sizeof(bmp_file_header),SEEK_SET);
	bmp_col[3] = 0;
	palette = pal;
	for (i = 0; i < min(256,colors); i++) {
		bmp_col[2] = (*palette)[i].r;
		bmp_col[1] = (*palette)[i].g;
		bmp_col[0] = (*palette)[i].b;
		if (fwrite(bmp_col,4,1,the_file) <= 0) file_ok = FALSE;
	}
	fseek(the_file,0,SEEK_END);
}

int bmp_file_saver::put_next_line(BYTE *line) {

	if ((line_n-- < 0) || (fwrite(line,bytesperline,1,the_file) <= 0))
		return(-1);
	return(1);
}

/*************************** GIF **********************/
/*************************** GIF **********************/
/*************************** GIF **********************/

gif_file::gif_file(char *fname) : graph_file_reader(fname) {
	int i;
	BYTE buff,sw;
	prefixcode = NULL;
	secondbyte = NULL;
	decodestack = NULL;
	buffer = NULL;

	format_file_error = 1;
	if  (!file_ok) return;	// return if constructor can't open the file
	init_buffer(FALSE);
	buffer = new BYTE[256];

	file_ok = FALSE;
	format_file_error = 3;

	i = fread(&header, sizeof(header), 1, the_file);	// get file header
	if ((i<=0) || (memcmp(&header.signature,GIFSIGNATURE,3))) return; //return if file is not GIF

	fread(&scrdesc, sizeof(scrdesc), 1, the_file);	// get screen descriptor
	height = scrdesc.height;
	width = scrdesc.width;

	line2standart = col2562byte; //just move
	switch(scrdesc.sizeofGCT) {
		case 7: source_type = COLOR256IMG; colors = 256; break;
		case 3: source_type = COLOR16IMG;
			source_ext_type = COLOR16_1PIXEL;
			colors = 16;
			break;
		case 0: source_type = MONOIMG;
			source_ext_type = MONOGIF;
			colors = 2;
			line2standart = monoGIF2mono;
			break;
		}
				// get global palette
	if (scrdesc.GCTflag) fread(palette, (1 << (scrdesc.sizeofGCT+1))*
						sizeof(colortype), 1, the_file);

	fread(&buff, 1, 1, the_file);
	if (buff != IMAGE_DESCRIPTOR) return; // operait only with image descriptor record

	fread(&imagedesc, sizeof(imagedesc), 1, the_file);	// get image descriptor record
	if (imagedesc.LCTflag) fread(palette, 1 << (imagedesc.sizeofLCT+1), 1, the_file);	// get palette

	for(i=0;i<colors;i++) {
		sw = palette[i].b;
		palette[i].b = palette[i].r;
		palette[i].r = sw;
	}

	fread(&LZW_code_size, 1, 1, the_file);	// get LZW code size
	LZW_code_size++;
	data_begin = ftell(the_file);

	prefixcode = (WORD*)malloc(LZW_TABLE_SIZE*sizeof(WORD));
	secondbyte = (BYTE*)malloc(LZW_TABLE_SIZE);
	decodestack = (BYTE*)malloc(LZW_TABLE_SIZE);
	source_line = new BYTE[width*4];
		// we'll finish readed block in decoding
		// so we need some exstra space in source_line
	if (source_type == MONOIMG) standart_line = new BYTE[width/8+1];
	else standart_line = new BYTE[width];
	if (prefixcode == NULL || secondbyte == NULL ||
	    decodestack == NULL	|| source_line == NULL) return;
	bytesperline = width;

	seek_start();
	file_ok = TRUE;
	format_file_error = 0;
}

WORD gif_file::getcode() {
	WORD code;
	while (bit_count < num_bits) {
		bit_buffer |= (DWORD) buffer[counter] << bit_count;
		bit_count += 8;
		if (++counter == blocksize) {
			fread(&blocksize, 1, 1, the_file);
			fread(buffer, blocksize, 1 ,the_file);
			counter = 0;
		}
	}
	code = bit_buffer & max_code;
	bit_buffer >>= num_bits;
	bit_count -= num_bits;
	return(code);
}

BYTE* gif_file::decode(BYTE *stack_bottom, WORD code){
	while(code > (clear_code-1)) {
		*stack_bottom++ = secondbyte[code];
		code = prefixcode[code];
	}
	*stack_bottom=code;
	return(stack_bottom);
}

int gif_file::get_next_line() {
	if(line_n < 0 || line_n >= height) return(-1);
	if (buff_offset > width)
		memmove(source_line,&source_line[width],width);
	else {
		memmove(source_line,&source_line[width],buff_offset);

		while(((buff_offset < width) &&
				(new_code=getcode()) != end_of_info)) {

			if (new_code == clear_code) {	// reset all decompression parameters
				clear_flag = TRUE;
				num_bits = LZW_code_size;
				next_code = first_code;
				max_code = (1L << LZW_code_size) - 1;
				continue;
			}
			if (clear_flag) {
				clear_flag = FALSE;
				source_line[buff_offset++] = old_code = pixelvalue = new_code;
				continue;
			}
			if (new_code >= next_code) {
				*decodestack = pixelvalue;
				stack_top = decode(decodestack+1,old_code);
			}
			else stack_top = decode(decodestack,new_code);

			pixelvalue = *stack_top;              // output decoded stack_top in reverse
			while (stack_top >= decodestack)
				source_line[buff_offset++] = *stack_top--;

			if (next_code <= max_code) {      // add to stack_top table if not full
				prefixcode[next_code] = old_code;
				secondbyte[next_code] = pixelvalue;
				if ((++next_code > max_code) && (num_bits < LZW_BITS))
					max_code = (1 << ++num_bits) - 1;
			}
			old_code = new_code;
		}
	}
	buff_offset -= width;
	line_n++;
	return(0);
}

void gif_file::seek_start() {
	graph_file_reader::seek_start();

	buff_offset = counter = bit_count = bit_buffer = 0;
	clear_flag = TRUE;
	clear_code = 1 << (LZW_code_size-1);
	end_of_info = clear_code + 1;
	first_code = clear_code + 2;
	next_code = first_code;
	max_code = (1 << LZW_code_size) - 1;
	num_bits = LZW_code_size;

	fread(&blocksize, 1, 1, the_file);
	fread(buffer, blocksize, 1 ,the_file);
}

gif_file::~gif_file() {
	if (prefixcode) free(prefixcode);
	if (secondbyte) free(secondbyte);
	if (decodestack) free(decodestack);
	if (source_line) delete source_line;
	if (standart_line) delete standart_line;
	if (buffer) delete buffer;
}

gif_file_saver::gif_file_saver(char *fname, image_type dest_type_,
			int width_,int height_,BGRpalette* pal) :
	graph_file_saver(fname,FALSE,
			dest_type_,NO_EXTENTION, //dest_ext_type_,
			width_,height_,pal) {

	prefixcode = NULL;
	secondbyte = NULL;
	code_value = NULL;
	buffer = NULL;

	line2standart = col2562byte; //just move
	switch  (dest_type) {
	case GRAY256IMG:  dest_ext_type = GRAY_PALETTE;
			 break;
	case TRUECOLORIMG: dest_type = COLOR256IMG;
			 break;
	case COLOR16IMG: dest_ext_type = COLOR16_1PIXEL;
			 break;
	case MONOIMG: dest_ext_type = MONOGIF;
		      line2standart = mono2GIFmono;
		      break;

	default: dest_ext_type =  NO_EXTENTION;
	}

	BYTE buff;

	format_file_error = 1;
	if  (!file_ok) return;	// return if constructor can't open the file

	format_file_error = 3;
	file_ok = FALSE;
	memmove(header.signature,GIFSIGNATURE,3);
	memmove(header.version,"87a",3);
	memset(&imagedesc,0,sizeof(imagedesc));
	memset(&scrdesc,0,sizeof(scrdesc));
	imagedesc.height = scrdesc.height = height;
	imagedesc.width = scrdesc.width = width;
	scrdesc.GCTflag = 1;

	switch(dest_type) {
	case MONOIMG: LZW_code_size = 2; break;

	case COLOR16IMG:
			LZW_code_size = 4;
			scrdesc.sizeofGCT = scrdesc.colorres = 3;
			break;
	case GRAY256IMG:
	case COLOR256IMG:
			LZW_code_size = 8;
			scrdesc.sizeofGCT = scrdesc.colorres = 7;
			break;
	}
	colors = 1 << (scrdesc.sizeofGCT + 1);

	fwrite(&header,sizeof(header),1,the_file);	// save GIF header
	fwrite(&scrdesc,sizeof(scrdesc),1,the_file);	//  save screen descriptor
	update_palette(pal);

	buff = IMAGE_DESCRIPTOR;
	fwrite(&buff,1,1,the_file);	// save IMAGE_DESCRIPTOR flag
	fwrite(&imagedesc,sizeof(imagedesc),1,the_file); // save image descriptor
	fwrite(&LZW_code_size,1,1,the_file);	// save LZW code size
	LZW_code_size++;

	prefixcode = (WORD*)malloc(HASH_SIZE*sizeof(WORD));
	code_value = (int*)malloc(HASH_SIZE*sizeof(int));
	secondbyte = (BYTE*)malloc(HASH_SIZE);
	buffer = new BYTE[256];

	if (prefixcode == NULL || secondbyte == NULL ||
	    code_value == NULL || buffer == NULL) return;

	memset(code_value,0xFF,HASH_SIZE*2); // initialize code table

	buff_offset = bit_count = bit_buffer = 0;
	clear_code = 1 << (LZW_code_size-1);
	end_of_info = clear_code + 1;
	first_code = clear_code + 2;
	next_code = first_code;
	max_code = (1 << LZW_code_size) - 1;
	num_bits = LZW_code_size;
	first_byte = TRUE;

	bytesperline = width;

	format_file_error = 0;
	file_ok = TRUE;
}

void gif_file_saver::update_palette(BGRpalette *pal) {
	VGApalette pall;
	int i;

	for(i = 0;i < colors;i++) {		// convert BGR palette to RGB palette
		pall[i].r = (*pal)[i].r;
		pall[i].g = (*pal)[i].g;
		pall[i].b = (*pal)[i].b;
	}
	fseek(the_file,sizeof(header)+sizeof(scrdesc),SEEK_SET);

	fwrite(pall,colors*sizeof(colortype),1,the_file);	// save global palette
	fseek(the_file,0,SEEK_END);
	palette = pal;
}

int gif_file_saver::put_next_line(BYTE *line) {
	WORD i=0,index;
	BYTE pixelvalue;

	if (line_n >= height) return(-1);
	if(first_byte) {
		code = line[0];			// get first pixel (code)
		putcode(clear_code);
		first_byte = FALSE;
		i++;
	}

	for(i=i;i<width;i++) {
		pixelvalue = line[i];		// get next pixel
		index = find_match(code,pixelvalue);

		if(code_value[index] != -1) code=code_value[index];
		else {
			code_value[index] = next_code;
			prefixcode[index] = code;
			secondbyte[index] = pixelvalue;
			putcode(code);		// send to stream current code
			code = pixelvalue;
			if (next_code++ > max_code)       // is table full
				if ( num_bits < LZW_BITS)
					max_code = (1 << ++num_bits) - 1; // increment code size
				else {
					putcode(clear_code);
					num_bits = LZW_code_size;
					next_code = first_code;
					max_code = (1 << num_bits) - 1;
					// initialize code table
					memset(code_value,0xFF,HASH_SIZE*2);
				}
		}
	}
	line_n++;
	if(line_n == height) {
		if (next_code == max_code)	// Handles special case for bit
			++num_bits;		// increment on EOF
		putcode(code);   		// output the last code
		putcode(end_of_info);
		putcode(0);
		if(buff_offset != 0) {
			fwrite(&buff_offset,1,1,the_file);
			fwrite(buffer,buff_offset,1,the_file);
		}
		buff_offset = 0x3B00;		// Zero length block & Trailer
		fwrite(&buff_offset,2,1,the_file);
	}
	return(0);
}

// output a variable length code
void gif_file_saver::putcode(WORD code) {
	bit_buffer |= (DWORD) code << bit_count;
	bit_count += num_bits;
	while (bit_count >= 8) {
		buffer[buff_offset] = bit_buffer & 0xFF;
		if(++buff_offset == 255) {
			fwrite(&buff_offset,1,1,the_file);
			fwrite(buffer,buff_offset,1,the_file);
			buff_offset = 0;
		}
		bit_buffer >>= 8;
		bit_count -= 8;
	}
}

// this is the hashing routine
int gif_file_saver::find_match(int hash_prefix,WORD hash_symbol) {
	int index,offset;

	index = (hash_symbol << LZW_BITS - 8) | hash_prefix;

	if(index == 0) offset = 1;
	else offset = HASH_SIZE - index;

	while(TRUE) {
		if(code_value[index] == -1) return(index);
		if(prefixcode[index] == hash_prefix &&
					secondbyte[index] == hash_symbol)
			return(index);
		index -= offset;
		if (index < 0) index += HASH_SIZE;
	}
}

gif_file_saver::~gif_file_saver() {
	free(prefixcode);
	free(code_value);
	free(secondbyte);
	delete buffer;
}

/*************************** PCX **********************/
/*************************** PCX **********************/
/*************************** PCX **********************/

pcx_file::pcx_file(char *fname) : graph_file_reader(fname) {
	int i;
	BYTE sw;

	format_file_error = 1;
	if  (!file_ok) return;	// return if constructor can't open the file

	format_file_error = 3;
	init_buffer(TRUE);
	file_ok = FALSE;
	data_begin = sizeof(header);

	i = fread(&header,sizeof(pcx_header),1,the_file);	// get file header
	if (i <= 0 || header.signature != PCXSIGNATURE ||
			header.version < 5) return; //return if file is not PCX

	height = header.y2 - header.y1 + 1;
	width = header.x2 - header.x1 + 1;

	line2standart = col2562byte; //just move
	switch(header.bitsperpixel+header.planes) {
		case 11: source_type = TRUECOLORIMG;
			colors = 1L << 24;
			source_ext_type = RGBPLANE24;
			line2standart = RGBplanes2RGBlines;
			break;
		case 9: {
			source_type = COLOR256IMG; colors = 256;
			fseek(the_file, 0, SEEK_END);
			fseek(the_file,-(long)sizeof(palette),SEEK_END);
			fread(palette, sizeof(palette), 1, the_file);	// get VGA palette
			fseek(the_file,sizeof(pcx_header),SEEK_SET);
			}
			break;
		case 5: {
			source_type = COLOR16IMG; colors = 16;
			memmove(palette,header.palette,16*sizeof(colortype));
			if(header.bitsperpixel == 4) {
				source_ext_type = COLOR16_2PIXEL;
				line2standart = col16A2col16B;
				}
			else {
				source_ext_type = COLOR16_PLANE;
				line2standart = col16plane2col16B;
			     }
			}
			break;
		case 2: source_type = MONOIMG;
			line2standart = mono2mono;
			colors = 2;
			break;
		default : return;
		}
	if(source_type != TRUECOLORIMG)
		for(i = 0;i < colors;i++) {		// convert BGR palette to RGB palette
			sw = palette[i].r;
			palette[i].r = palette[i].b;
			palette[i].b = sw;
		}
	bytesperline = (width*header.planes*header.bitsperpixel+7)/8;
	real_len = (width*header.bitsperpixel+7)/8;
	source_line = new BYTE[bytesperline+header.planes];

	if   (source_type == COLOR16IMG) standart_line = new BYTE[2*bytesperline];
	else standart_line = new BYTE[bytesperline];

	file_ok = TRUE;
	format_file_error = 0;
	seek_start();
}

int pcx_file::get_next_line() {
	register char a,k;
	register int counter;
	int i,len = header.bytesperline;
	BOOL flag;
	BYTE *s_l = source_line, *buf;

	if(line_n < 0 || line_n >= height) return(-1);

	fresh_buffer();

	buf = &buffer[buff_offset];
	for(i=0;i<header.planes;i++) {
		flag = TRUE; counter = 0;
		while(flag)
			if(prev_c != 0) {
				for(k=prev_c;k<rep;k++) {
					*(s_l++) = datab;
					if(++counter == len) {flag = FALSE;break;}
				}
				prev_c = 0;
			}
			else {
			a = *(buf++); buff_offset++;
			if((a & 0xC0) == 0xC0) {
				datab = *(buf++); buff_offset++;
				rep = a & 0x3F;
				for(k=0;k<rep;k++) {
					*(s_l++) = datab;
					if(++counter == len) {
						prev_c = k+1;
						flag = FALSE;
						break;
					}
				}
			}
			else { *(s_l++) = a; if(++counter == len) flag = FALSE;}
		}
		if(len != real_len) s_l--;	// skip left-hand byte
	}
	line_n++;
	return(0);
}

void pcx_file::seek_start() {
	prev_c = 0;
	graph_file_reader::seek_start();
}

pcx_file::~pcx_file() {
	if (source_line) delete source_line;
	if (standart_line) delete standart_line;
}

pcx_file_saver::pcx_file_saver(char *fname, image_type dest_type_,
			image_extention_type dest_ext_type_,
			int width_,int height_,BGRpalette* pal) :
		graph_file_saver(fname,TRUE,dest_type_,dest_ext_type_,
					width_,height_,pal) {
	int i;

	format_file_error = 1;
	if  (!file_ok) return;	// return if constructor can't open the file

	file_ok = FALSE;
	format_file_error = 3;
	memset(&header,0,sizeof(header));

	header.x2 = width - 1;
	header.y2 = height - 1;
	header.signature = PCXSIGNATURE;
	header.version = 5;
	header.palinterp = header.encoding = 1;

	line2standart = col2562byte; //just move
	switch(dest_type) {
	case MONOIMG:
			for(i=0;i<2;i++) {
				header.palette[i].r = (*palette)[i].r;
				header.palette[i].g = (*palette)[i].g;
				header.palette[i].b = (*palette)[i].b;
			}
			header.planes = header.bitsperpixel = 1;
			line2standart = mono2mono;
			break;
	case COLOR16IMG:
			for(i=0;i<16;i++) {
				header.palette[i].r = (*palette)[i].r;
				header.palette[i].g = (*palette)[i].g;
				header.palette[i].b = (*palette)[i].b;
			}
			if(dest_ext_type == COLOR16_2PIXEL) {
				header.bitsperpixel = 4;
				header.planes = 1;
				line2standart = col16B2col16A;
			}
			else {
				header.bitsperpixel = 1;
				header.planes = 4;
				line2standart = col16B2col16plane;
			}
			break;
	case GRAY256IMG : dest_ext_type = GRAY_PALETTE;
	case COLOR256IMG:
			header.bitsperpixel = 8;
			header.planes = 1;
			break;
	case TRUECOLORIMG:
			header.bitsperpixel = 8;
			header.planes = 3;
			line2standart = RGBplanes2RGBlines;
			break;
	default : return;
	}

	bytesperline = header.bytesperline = (width*header.bitsperpixel + 7) / 8;
	bytesperline *= header.planes;
	fwrite(&header,sizeof(header),1,the_file);
	file_ok = TRUE;
	format_file_error = 0;
}

void pcx_file_saver::update_palette(BGRpalette *pal) {
	int i;
	if (dest_type == TRUECOLORIMG) return;
	if (dest_type == COLOR16IMG || dest_type == MONOIMG) {
		for(i=0;i<16;i++) {
			header.palette[i].r = (*pal)[i].r;
			header.palette[i].g = (*pal)[i].g;
			header.palette[i].b = (*pal)[i].b;
		}
	fseek(the_file,0,SEEK_SET);
	fwrite(&header,sizeof(header),1,the_file);
	fseek(the_file,0,SEEK_END);
	}
	palette = pal;
}

int pcx_file_saver::put_next_line(BYTE *line) {
	int i,counter=0;
	BYTE c=0;
	BYTE *buf;
	colortype *pal;

	buf = &buffer[buff_offset];
	while(counter < bytesperline) {
			if(line[counter++] == line[counter]) {
				if(++c == 0x3F || counter == bytesperline) {
					*(buf++) = c | 0xC0;
					*(buf++) = line[counter];
					buff_offset += 2;
					c=0;
				}
			}
			else
				if(c != 0) {
					*(buf++) = (c+1) | 0xC0;
					*(buf++) = line[counter-1];
					buff_offset += 2;
					c = 0;
				}
				else {
					if((line[counter-1] & 0xC0) == 0xC0) {
						*(buf++) = 0xC1;
						buff_offset++;
					}
					*(buf++) = line[counter-1]; buff_offset++;
				}
		}

	line_n++;
	if(buff_offset > (BUFF_SIZE - bytesperline*2) || line_n == height) {
		flush_buffer();
		if((line_n == height) &&
			(dest_type == GRAY256IMG || dest_type == COLOR256IMG)) {
			c = 0x0C;
			fwrite(&c,1,1,the_file);
			pal = new colortype [256];
			for(i=0;i<256;i++) {
				pal[i].r = (*palette)[i].r;
				pal[i].g = (*palette)[i].g;
				pal[i].b = (*palette)[i].b;
			}
			fwrite(pal,sizeof(VGApalette),1,the_file);
			delete pal;
		}
	}
	return(0);
}


char *format_file_error_report() {
	return format_file_error_messages[format_file_error];
	}


graph_file_reader *open_image_file(char *fname) { // returns null if error
	char drive[MAXDRIVE];
	char dir[MAXDIR];
	char file[MAXFILE];
	char ext[MAXEXT];
	graph_file_reader *grf_r = NULL;

	format_file_error = 0;
	fnsplit(fname,drive,dir,file,ext);  // split the string to separate elems
	strlwr(&ext[1]);
	if(!memcmp(&ext[1],"pcx",3)) grf_r  = new pcx_file(fname); else
	if(!memcmp(&ext[1],"gif",3)) grf_r  = new gif_file(fname); else
	if(!memcmp(&ext[1],"bmp",3) ||
		!memcmp(&ext[1],"dib",3)) grf_r  = new bmp_file(fname); else
	if(!memcmp(&ext[1],"tga",3)) grf_r  = new targa_file(fname);
	else {
		format_file_error = 2;
		return(NULL);
	     }

	if (!grf_r-> file_ok) {
		if (format_file_error == 0) format_file_error = 4;
		delete grf_r;
		grf_r = NULL;
		}
	return(grf_r);
}

graph_file_saver *open_saver_file(char *fname,
		image_type dest_type,
		image_extention_type dest_ext_type,
		int width, int height,
		BGRpalette* palptr,
		BOOL top_to_bottom, // if possible
		BOOL left_to_right) { // returns null if error

	char drive[MAXDRIVE];
	char dir[MAXDIR];
	char file[MAXFILE];
	char ext[MAXEXT];

	graph_file_saver *grf_s;

	format_file_error = 0;
	fnsplit(fname,drive,dir,file,ext);  // split the string to separate elems
	strlwr(&ext[1]);

	if(!memcmp(&ext[1],"pcx",3))
		grf_s = new pcx_file_saver(fname,
			dest_type, dest_ext_type,
			width,height,	palptr);

	else if(!memcmp(&ext[1],"gif",3))
		grf_s = new gif_file_saver(fname,dest_type,
			width, height, palptr);

	else if(!memcmp(&ext[1],"bmp",3) || !memcmp(&ext[1],"dib",3))
		grf_s = new bmp_file_saver(fname,dest_type,
			 width, height, palptr);

	else if(!memcmp(&ext[1],"tga",3))
		grf_s = new targa_file_saver(fname,
			dest_type, dest_ext_type,
			width, height,
			palptr,TRUE,
			top_to_bottom, left_to_right);
	else {
		format_file_error = 2;
		return(NULL);
	     }

	if(!grf_s->file_ok) {
		if (format_file_error == 0) format_file_error = 4;
		delete grf_s;
		grf_s = NULL;
	}
	return(grf_s);
}

void choosevideomode(graph_file_reader *grf_r) {
	if(modelist.numberofmodes == 1) {
		printf("error graphics initializing.\n");
		exit(0);
	}
	switchmode(VGA256);
	if(grf_r->height <= 200 && grf_r->width <= 320) return;
	switchmode(SVGA640x480x256);
	if(grf_r->height <= 480 && grf_r->width <= 640) return;
	/*
	switchmode(SVGA800x600x256);
	if(grf_r->height <= 800 && grf_r->width <= 600) return;

	switchmode(SVGA1024x768x256);
	*/
}

void setpalette_by_BGR(BGRpalette pal) {
	VGApalette vpal;
	int i;
	for (i = 255; i >= 0; i--) {
		vpal[i].r = pal[i].r >> 2;
		vpal[i].g = pal[i].g >> 2;
		vpal[i].b = pal[i].b >> 2;
		}
	setDACblock(0,256,vpal);
}

const rbits = 3;
const gbits = 3;
const bbits = 2;
// init simple standard palette for 256 colors mode
void stand_palette(VGApalette &pal) {
	int i = 0,r,g,b;
	for ( r = 0; r < (1<<rbits); r++)
		for ( g = 0; g < (1<<gbits); g++)
			for ( b = 0; b < (1<<bbits); b++){
				pal[i].r = r << (6 - rbits);
				pal[i].g = g << (6 - gbits);
				pal[i++].b = b << (6 - bbits);
				}
      setDACblock(0,256,pal);
}

void put_TC_row(BYTE *rp, BYTE *gp, BYTE *bp,
		int x,int y,int xcount) {
	int i,r,g,b,c;
	BYTE *line,*l;
	if  (xcount <= 0) return;
	line = new BYTE[xcount];
	l = line;
	for ( i = 0; i < xcount; i++) {
		c =  (*(rp++) >> (8-rbits)) << (gbits+bbits);
		c += (*(gp++) >> (8-gbits)) << bbits;
		c += *(bp++) >> (8-bbits);
		*(l++) = c;
	}
	putrow(rowptr(line),x,y,xcount);
	delete line;
}
