/* color mapping for selected palette */

#include <mem.h>
#include <alloc.h>
#include "common.h"
#include "graph.h"
#include "palette.h"
#include "mapper2.h"

/*
 * These routines are concerned with the time-critical task of mapping input
 * colors to the nearest color in the selected colormap.
/
/*
 * The next three routines implement inverse colormap filling.  They could
 * all be folded into one big routine, but splitting them up this way saves
 * some stack space (the mindist[] and bestdist[] arrays need not coexist)
 * and may allow some compilers to produce better code by registerizing more
 * inner-loop variables.
 */

int color_mapper_palette::find_nearby_colors (int minc0, int minc1, int minc2,BYTE colorlist[])
/* Locate the colormap entries close enough to an update box to be candidates
 * for the nearest entry to some cell(s) in the update box.  The update box
 * is specified by the center coordinates of its first cell.  The number of
 * candidate colormap entries is returned, and their colormap indexes are
 * placed in colorlist[].
 * This routine uses Heckbert's "locally sorted search" criterion to select
 * the colors that need further consideration.
 */
{
  register int x;
  int numcolors = actual_number_of_colors;
  int maxc0, maxc1, maxc2;
  int centerc0, centerc1, centerc2;
  int i, ncolors;
  long minmaxdist, min_dist, max_dist, tdist;
  long mindist[MAXNUMCOLORS];	/* min distance to colormap entry i */
  BGRcolortype  color;
  /* Compute true coordinates of update box's upper corner and center.
   * Actually we compute the coordinates of the center of the upper-corner
   * map cell, which are the upper bounds of the volume we care about.
   * Note that since ">>" rounds down, the "center" values may be closer to
   * min than to max; hence comparisons to them must be "<=", not "<".
   */
  maxc0 = minc0 + ((1 << BOX_R_SHIFT) - (1 << MAP_R_SHIFT));
  centerc0 = (minc0 + maxc0) >> 1;
  maxc1 = minc1 + ((1 << BOX_G_SHIFT) - (1 << MAP_G_SHIFT));
  centerc1 = (minc1 + maxc1) >> 1;
  maxc2 = minc2 + ((1 << BOX_B_SHIFT) - (1 << MAP_B_SHIFT));
  centerc2 = (minc2 + maxc2) >> 1;

  /* For each color in colormap, find:
   *  1. its minimum squared-distance to any point in the update box
   *     (zero if color is within update box);
   *  2. its maximum squared-distance to any point in the update box.
   * Both of these can be found by considering only the corners of the box.
   * We save the minimum distance for each color in mindist[];
   * only the smallest maximum distance is of interest.
   * Note we have to scale coordinates to get correct distance in scaled space.
   */
  minmaxdist = 0x7FFFFFFFL;

  for (i = 0; i < numcolors; i++) {
    /* We compute the squared-c0-distance term, then add in the other two. */
    color = my_colormap[i];
    x = color.r;

    if (x < minc0) {
      tdist = (x - minc0) * RSCALE >> 4;
      min_dist = tdist*tdist;
      tdist = (x - maxc0) * RSCALE << 4;
      max_dist = tdist*tdist;
    } else if (x > maxc0) {
      tdist = (x - maxc0) * RSCALE >>4;
      min_dist = tdist*tdist;
      tdist = (x - minc0) * RSCALE >>4;
      max_dist = tdist*tdist;
    } else {
      /* within cell range so no contribution to min_dist */
      min_dist = 0;
      if (x <= centerc0) {
	tdist = (x - maxc0) * RSCALE >>4;
	max_dist = tdist*tdist;
      } else {
	tdist = (x - minc0) * RSCALE >>4;
	max_dist = tdist*tdist;
      }
    }

    x = color.g;
    if (x < minc1) {
      tdist = (x - minc1) * GSCALE >>4;
      min_dist += tdist*tdist;
      tdist = (x - maxc1) * GSCALE >>4;
      max_dist += tdist*tdist;
    } else if (x > maxc1) {
      tdist = (x - maxc1) * GSCALE >>4;
      min_dist += tdist*tdist;
      tdist = (x - minc1) * GSCALE >>4;
      max_dist += tdist*tdist;
    } else {
      /* within cell range so no contribution to min_dist */
      if (x <= centerc1) {
	tdist = (x - maxc1) * GSCALE >>4;
	max_dist += tdist*tdist;
      } else {
	tdist = (x - minc1) * GSCALE >>4;
	max_dist += tdist*tdist;
      }
    }

    x = color.b;
    if (x < minc2) {
      tdist = x - minc2;
      min_dist += tdist*tdist;
      tdist = x - maxc2;
      max_dist += tdist*tdist;
    } else if (x > maxc2) {
      tdist = x - maxc2;
      min_dist += tdist*tdist;
      tdist = x - minc2;
      max_dist += tdist*tdist;
    } else {
      /* within cell range so no contribution to min_dist */
      if (x <= centerc2) {
	tdist = x - maxc2;
	max_dist += tdist*tdist;
      } else {
	tdist = x - minc2;
	max_dist += tdist*tdist;
      }
    }

    mindist[i] = min_dist;	/* save away the results */
    if (max_dist < minmaxdist)
      minmaxdist = max_dist;
  }

  /* Now we know that no cell in the update box is more than minmaxdist
   * away from some colormap entry.  Therefore, only colors that are
   * within minmaxdist of some part of the box need be considered.
   */
  ncolors = 0;
  for (i = 0; i < numcolors; i++) {
    if (mindist[i] <= minmaxdist)
			colorlist[ncolors++] =  i;
  }
  return ncolors;
}


void color_mapper_palette::find_best_colors (int minc0, int minc1, int minc2,
		  int numcolors, BYTE colorlist[], BYTE bestcolor[])
/* Find the closest colormap entry for each cell in the update box,
 * given the list of candidate colors prepared by find_nearby_colors.
 * Return the indexes of the closest entries in the bestcolor[] array.
 * This routine uses Thomas' incremental distance calculation method to
 * find the distance from a colormap entry to successive cells in the box.
 */
{
  int ic0, ic1, ic2;
  int i, icolor;
  register long * bptr;	/* pointer into bestdist[] array */
  BYTE * cptr;		/* pointer into bestcolor[] array */
  long dist0, dist1;		/* initial distance values */
  register long dist2;		/* current distance in inner loop */
  long xx0, xx1;		/* distance increments */
  register long xx2;
  /* This array holds the distance to the nearest-so-far color for each cell */
  long bestdist[BOX_R_ELEMS * BOX_G_ELEMS * BOX_B_ELEMS];
  int inc0, inc1, inc2;
  /* Initialize best-distance for each cell of the update box */
  bptr = bestdist;

  for (i = BOX_R_ELEMS*BOX_G_ELEMS*BOX_B_ELEMS-1; i >= 0; i--)
    *bptr++ = 0x7FFFFFFFL;

  /* For each color selected by find_nearby_colors,
   * compute its distance to the center of each cell in the box.
   * If that's less than best-so-far, update best distance and color number.
   * Note we have to scale G,R to get correct distance in scaled space.
   */

  /* Nominal steps between cell centers ("x" in Thomas article) */
#define STEP_R  ((1 << MAP_R_SHIFT) * RSCALE >>4)
#define STEP_G  ((1 << MAP_G_SHIFT) * GSCALE >>4)
#define STEP_B  (1 << MAP_B_SHIFT)

  for (i = 0; i < numcolors; i++) {
    icolor = colorlist[i];
    /* Compute (square of) distance from minc0/c1/c2 to this color */
    inc0 = (minc0 - (int) my_colormap[icolor].r)* RSCALE >>4;
    dist0 = (long)inc0*inc0;
    inc1 = (minc1 - (int) my_colormap[icolor].g)* GSCALE >>4;
    dist0 += (long)inc1*inc1;
    inc2 = minc2 - (int) my_colormap[icolor].b;
    dist0 += (long)inc2*inc2;
    /* Form the initial difference increments */
    inc0 = inc0 * (2 * STEP_R) + STEP_R * STEP_R;
    inc1 = inc1 * (2 * STEP_G) + STEP_G * STEP_G;
    inc2 = inc2 * (2 * STEP_B) + STEP_B * STEP_B;
    /* Now loop over all cells in box, updating distance per Thomas method */
    bptr = bestdist;
    cptr = bestcolor;
    xx0 = inc0;
    for (ic0 = BOX_R_ELEMS-1; ic0 >= 0; ic0--) {
      dist1 = dist0;
      xx1 = inc1;
      for (ic1 = BOX_G_ELEMS-1; ic1 >= 0; ic1--) {
	dist2 = dist1;
	xx2 = inc2;
	for (ic2 = BOX_B_ELEMS-1; ic2 >= 0; ic2--) {
	  if (dist2 < *bptr) {
	    *bptr = dist2;
	    *cptr =  icolor;
	  }
	  dist2 += xx2;
	  xx2 +=  2 * STEP_B * STEP_B;
	  bptr++;
	  cptr++;
	}
	dist1 += xx1;
	xx1 += 2 * STEP_G * STEP_G;
      }
      dist0 += xx0;
      xx0 += 2 * STEP_R * STEP_R;
    }
  }
}


void color_mapper_palette::fill_inverse_cmap (int c0, int c1, int c2) {
/* Fill the inverse-colormap entries in the update box that contains */
/* map cell c0/c1/c2.  (Only that one cell MUST be filled, but */
/* we can fill as many others as we wish.) */

  int minc0, minc1, minc2;	/* lower left corner of update box */
  int ic0, ic1, ic2;
  register BYTE * cptr;	/* pointer into bestcolor[] array */
  register mapptr cachep;	/* pointer into main cache array */
  /* This array lists the candidate colormap indexes. */
  BYTE colorlist[256];
  int numcolors;		/* number of candidate colors */
  /* This array holds the actually closest colormap index for each cell. */
  BYTE bestcolor[BOX_R_ELEMS * BOX_G_ELEMS * BOX_B_ELEMS];

  /* Convert cell coordinates to update box ID */
  c0 >>= BOX_R_LOG;
  c1 >>= BOX_G_LOG;
  c2 >>= BOX_B_LOG;

  /* Compute true coordinates of update box's origin corner.
   * Actually we compute the coordinates of the center of the corner
   * map cell, which are the lower bounds of the volume we care about.
   */
  minc0 = (c0 << BOX_R_SHIFT) + ((1 << MAP_R_SHIFT) >> 1);
  minc1 = (c1 << BOX_G_SHIFT) + ((1 << MAP_G_SHIFT) >> 1);
  minc2 = (c2 << BOX_B_SHIFT) + ((1 << MAP_B_SHIFT) >> 1);

  /* Determine which colormap entries are close enough to be candidates
   * for the nearest entry to some cell in the update box.
   */
  numcolors = find_nearby_colors(minc0, minc1, minc2, colorlist);

  /* Determine the actually nearest colors. */
  find_best_colors(minc0, minc1, minc2, numcolors, colorlist,  bestcolor);

  /* Save the best color numbers (plus 1) in the main cache array */
	c0 <<= BOX_R_LOG;		/* convert ID back to base cell indexes */
	c1 <<= BOX_G_LOG;
	c2 <<= BOX_B_LOG;
	cptr = bestcolor;
	for (ic0 = 0; ic0 < BOX_R_ELEMS; ic0++) {
		for (ic1 = 0; ic1 < BOX_G_ELEMS; ic1++) {
			// cachep = & map[c0+ic0][c1+ic1][c2];
			cachep = & map[c0+ic0][((c1+ic1) << MAP_B_BITS)+c2];
			for (ic2 = 0; ic2 < BOX_B_ELEMS; ic2++)
			 *cachep++ = (mapcell) (BYTE(*cptr++) + 1);
		}
	}
}


/*
 * These routines perform second-pass scanning of the image: map each pixel to
 * the proper colormap index, and output the indexes to the output file.
 *
 * output_workspace is a one-component array of pixel dimensions at least
 * as large as the input image strip; it can be used to hold the converted
 * pixels' colormap indexes.
 */


void color_mapper_palette::process_line(BYTE *R, BYTE *G, BYTE *B, BYTE *out_line){

/* This version performs no dithering */

  BYTE *ptr0, *ptr1, *ptr2, *outptr;
  register mapptr cachep;
  register int c0, c1, c2;
  int col; // originally long type was used, I don't beleive in
	   // lines 32000 pixels long

/* Convert data to colormap indexes, which we save in out_line */
	ptr0 = R;
	ptr1 = G;
	ptr2 = B;
	outptr = out_line;
	for (col = width; col > 0; col--) {
		/* get pixel value and index into the cache */
		c0 = (*ptr0++) >> MAP_R_SHIFT;
		c1 = (*ptr1++) >> MAP_G_SHIFT;
		c2 = (*ptr2++) >> MAP_B_SHIFT;
		//		cachep = & map[c0][c1][c2];
		cachep =  &map [c0] [(c1 << MAP_B_BITS) | c2 ];

		/* If we have not seen this color before, find nearest colormap entry */
		/* and update the cache */
		if (*cachep == 0) fill_inverse_cmap(c0,c1,c2);
		/* Now emit the colormap index for this cell */
		*outptr++ = (mapcell) (*cachep) - 1;
	}
}


/*
 * Initialize for two-pass color quantization.
 */

int color_mapper_palette::initmap() {
	int i;

	/* Allocate and zero the map */
	map_r_elements = MAP_R_ELEMS; // for destructor virtual
	/* I killed thoroughly distingued types of memory allocation - E.P.*/
	// map = new map2d[MAP_R_ELEMS];
	map = new mapptr[MAP_R_ELEMS];

	for (i = 0; i < MAP_R_ELEMS; i++) map[i] = NULL;

	for (i = 0; i < MAP_R_ELEMS; i++) {
	//		map[i] = (map2d) new mapcell[MAP_G_ELEMS][MAP_B_ELEMS];
		map[i] =  new mapcell[MAP_G_ELEMS*MAP_B_ELEMS];
		if (map[i] == NULL) return(-1);
		memset(map[i],0,MAP_G_ELEMS*MAP_B_ELEMS * sizeof(mapcell));
	}
	return(1);
}

color_mapper_palette::color_mapper_palette(int colors_,int width_,BGRpalette colormap) :
	color_mapper(colors_,width_,colormap) {
  initialized = FALSE;
  if (colors_ > MAXNUMCOLORS) return;
  if (this->initmap() > 0) initialized = TRUE;
}

color_mapper_palette::~color_mapper_palette() {
	int i;
	for (i = 0; i < MAP_R_ELEMS; i++)  if (map[i] != NULL) delete map[i];
	delete map;
}


// with dithering ***************************************************

color_mapper_palette_dith::color_mapper_palette_dith(int colors_,int width_,BGRpalette colormap)
	:color_mapper_palette(colors_,width_,colormap) {
	initialized = (prepare_dithering() > 0);
}


void color_mapper_palette_dith::process_line(BYTE *R, BYTE *G, BYTE *B, BYTE *out_line){
// dithering
  mapptr cachep;
  register int valR, valG, valB;
  int vr,vg,vb;
  int column; // originally long type was used, I don't beleive in
		// lines 32000 pixels long

  register int val;
  int two_val,i;
  int *swp;
  BYTE *convert_tab, *value_tab;
  register int pixcode;
  int dir; /* 1 for left-to-right, -1 for right-to-left */
  BGRcolortype color;
  int *thisrowerrR, *thisrowerrG, *thisrowerrB,
       *nextrowerrR, *nextrowerrG, *nextrowerrB;

  thisrowerrR = error[0] + 1;
  thisrowerrG = error[1] + 1;
  thisrowerrB = error[2] + 1;
  nextrowerrR = next_line_error[0] + width;
  nextrowerrG = next_line_error[1] + width;
  nextrowerrB = next_line_error[2] + width;

  if (on_odd_row) {
	/* work right to left in this row */
	dir = -1;
	R += width-1;
	G += width-1;
	B += width-1;
	out_line += width-1;
      } else {
	/* work left to right in this row */
	dir = 1;
      }

  for (i = 0; i < 3; i++)  next_line_error[i][width] = 0; /* need only initialize this one entry */

  for (column = width; column > 0; column--) {
	/* Get accumulated error for each component, round to integer.
	 * (error was scaled in 16 times)
	 * RIGHT_SHIFT rounds towards minus infinity, so adding 8 is correct
	 * for either sign of the error value.
	 */
//	 if (column == 300)
//	  column = 300	;

	 valR = *R + ((*thisrowerrR + 8) >> 4);
	 valG = *G + ((*thisrowerrG + 8) >> 4);
	 valB = *B + ((*thisrowerrB + 8) >> 4);

       /*
	 valR = *R; //  debug
	 valG = *G;
	 valB = *B;
       */
	/* Compute pixel value + error compensation,
	 *  Note max error value is +- MAXJSAMPLE.
	 */

	// find color index by R,G,B

	if hi(valR) {
		if (valR < 0) {vr = 0; valR = 0;}
		else if (valR > 255) {vr = 255 >> MAP_R_SHIFT; valR = 255;}
	}
	else vr = valR >>  MAP_R_SHIFT;

	if hi(valG) {
		if (valG < 0) { vg = 0; valG = 0;}
		else if (valG > 255) {vg = 255 >> MAP_G_SHIFT; valG = 255;}
	}
	else vg = valG >> MAP_G_SHIFT;

	if hi(valB) {
		if (valB < 0) {vb = 0; valB = 0;}
		else if (valB > 255) {vb = 255 >> MAP_B_SHIFT; valB = 255;}
	}
	else vb = valB >> MAP_B_SHIFT;

	// cachep = & map[vr][vg][vb];
	cachep = & map[vr][(vg << MAP_B_BITS) | vb];

	/* If we have not seen this color before, find nearest colormap entry */
	/* and update the cache */
	if (*cachep == 0) fill_inverse_cmap(vr,vg,vb);
	pixcode = *cachep-1;

	*out_line = pixcode;

	color = my_colormap[pixcode];
	/* Compute actual representation error at this pixel */
	/* Note: we can do this even though we don't yet have the final */
	/* value of pixcode, because the colormap is orthogonal. */
	valR -= color.r;
	/* Propagate error to (same component of) adjacent pixels */
	/* Remember that nextrowerr entries are in reverse order! */

	two_val = valR << 1;    // *2
	nextrowerrR[-1]  = valR; /* not +=, since not initialized yet */
	valR += two_val;		/* form error * 3 */
	nextrowerrR[ 1] += valR;
	valR += two_val;		/* form error * 5 */
	nextrowerrR[ 0] += valR;
	valR += two_val;		/* form error * 7 */
	thisrowerrR[ 1] += valR;

	valG -= color.g;
	two_val = valG << 1;    // *2
	nextrowerrG[-1]  = valG; /* not +=, since not initialized yet */
	valG += two_val;		/* form error * 3 */
	nextrowerrG[ 1] += valG;
	valG += two_val;		/* form error * 5 */
	nextrowerrG[ 0] += valG;
	valG += two_val;		/* form error * 7 */
	thisrowerrG[ 1] += valG;

	valB -= color.b;
	two_val = valB << 1;    // *2
	nextrowerrB[-1]  = valB; /* not +=, since not initialized yet */
	valB += two_val;		/* form error * 3 */
	nextrowerrB[ 1] += valB;
	valB += two_val;		/* form error * 5 */
	nextrowerrB[ 0] += valB;
	valB += two_val;		/* form error * 7 */
	thisrowerrB[ 1] += valB;


	R += dir;	/* advance input ptr to next column */
	G += dir;	/* advance input ptr to next column */
	B += dir;	/* advance input ptr to next column */
	out_line += dir;	/* advance output ptr to next column */
	thisrowerrR++;		/* cur-row error ptr advances to right */
	thisrowerrG++;
	thisrowerrB++;
	nextrowerrR--;		/* next-row error ptr advances to left */
	nextrowerrG--;
	nextrowerrB--;
      }

  for (i = 0; i < 3; i++) {
	swp = error[i];
	error[i] = next_line_error[i];          // swap pointers for future
	next_line_error[i] = swp;
	}
  on_odd_row = (on_odd_row ? FALSE : TRUE);
}




