/**************************************************************************
How to use Mandelbrot-like sets explorer.
Command line parameters:
EXPLORER {type of set} {input file} {output files base name} [iterations]

{type of set} can be:
	"m" for Mandelbrot,
	"j" for Julia,
	"dm" for Mandelbrot Dragon
	"dj" for Julia Dragon.

{input file} is name of a file created by the explorer.
It's a .TGA image with some extra information used by the explorer.

{output files base name} is a pathname (the file name part can be up to
6 characters). The explorer will save files adding 00, 01, 02, etc.
and the extension .TGA to the specified base name.

[iterations] are optional. Default is 64.

To create a new input file, run EXPLORER with "_new"
instead of input file and full pathname instead of an output base name,
and width and height of the image:
E.g.:

EXPLORER m _new c:\mynewest.tga 320 200

will create a 320x200 Mandelbrot set which you'll be able to explore
later.

When you start EXPLORER with the type as in the input file,
you see the navigation window which you can move around with arrow keys
(or the mouse). You can resize this window by holding down the Shift key
and moving arrow keys (or holding down left mouse button and moving
the mouse). When you press Enter, the navigation window is blown up to
the whole image size (recalculated). Alt-S saves the current image
to a file (base00.tga, base01.tga, etc), Esc exits the program.

If specify j type (Julia) with a Mandelbrot input image, then EXPLORER
displays a cross to select in which point of a Mandelbrot set to
calculate Julia set. You move the cross, press Enter, and then you
see the navigation window, go deeper, etc. (as described in the paragraph
above). Hitting Backspace will bring you back to the Mandelbrot set.
***********************************************************************/

#pragma -ml

#include <math.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <dir.h>

#include "common.h"
#include "mand_set.h"
#include "grfile.h"
#include "formats.h"
#include "keymouse.h"
#include "set_stuf.h"

graph_file_reader *grf_r;
mandelbrot_set	 *ms = NULL;
mset_descr descr = {MANDELBROT, -2.0, 1.2, -1.2, 1.2, 320, 200};
VGApalette palette;
BGRpalette pal;

void show_file(graph_file_reader *grf_r, int left, int top) {
	int i;

	grf_r -> seek_start();
	for (i = 0; i < grf_r -> height; i++) {
		grf_r->get_next_line();
		putrow((rowptr)grf_r->source_line,left,i+top,grf_r->width);
	}
};

// select point on pre-calculated image  - map of new images
BOOL mand_point(pointtype &p, RECT restrictions) { // uses grf_r
	BOOL ok = FALSE, finished = FALSE;
	WORD key;
	int v;
	double c_re, c_im;

	show_file(grf_r, ms->screen_left, ms->screen_top);
	do  {
		key = select_point(p, restrictions);
		finished = FALSE;  ok = TRUE;
		switch (key) {
		    case kwEsc:    ok = FALSE;
		    case kwEnter:  finished = TRUE; break;
		    }
	  } while (!finished);

      if (ok) {
	      v = p.x - ms->screen_left;
	      c_re = descr.x_min+ (descr.x_max - descr.x_min)*
					v / descr.screen_x;
	      v = p.y - ms->screen_top;
	      c_im = descr.y_min+ (descr.y_max - descr.y_min)*
					v / descr.screen_y;
	      ms -> set_c(c_re,c_im);
	      if (ms->my_type() == JULIA)
		      ms-> set_ranges(-1.5, 1.5, -1.2, 1.2);
	      else    ms-> set_ranges(-1.0, 2.0, -1.2, 1.2);

      }

      return(ok);
}

void type_error() {
	closegraph();
	printf("Type contradiction.\n");
	exit(0);
}

void main(int argc, char **argv) {
	int i,currline;
	RECT win = {{70,70},{120,120}},
	     restrictions = {{0,0},{199,199}};
	pointtype p = {100,100};
	BOOL ok = FALSE,
	     clipping,
	     int_type = FALSE,   // integer calculations
	     julia_like = FALSE, //
	     no_mand_map = FALSE; //
	WORD key;

	// int scrx = 200, scry = 200;
	char *infilename = "mand.tga";
	char outfilename[100] = "julia_.tga";
	int saveindex = 0;
	int iterno = 64;
	int repeater = 0;


	SET_TYPE set_type = MANDELBROT;

	char drive[MAXDRIVE];
	char dir[MAXDIR];
	char file[MAXFILE];
	char ext[MAXEXT];

	colortype colordef[8] = {
		{0,0,128},
		{0,0,255},
		{0,255,255},
		{0,255,0},
		{255,255,0},
		{255,0,0},
		{255,0,255},
		{255,255,255},
		};
	float c_re = 0.0, c_im = 0.0;

	init_palette(palette, 8, colordef);
	palette[0].r = 0;
	palette[0].g = 0;
	palette[0].b = 0;
	palette[255].r = 255;
	palette[255].g = 255;
	palette[255].b = 255;
	for (i = 0; i < 256; i++) {
		pal[i].r = palette[i].r;
		pal[i].g = palette[i].g;
		pal[i].b = palette[i].b;
		}

	for (i = 0; i < 256; i++) {
		palette[i].r >>= 2;
		palette[i].g >>= 2;
		palette[i].b >>= 2;
		}

	// process parameters
	if (argc > 1)  {
		if (!memicmp(argv[1],"M",2)) set_type = MANDELBROT;
		else if  (!memicmp(argv[1],"J",2))  set_type = JULIA;
		else if  (!memicmp(argv[1],"DM",3))  set_type = DRAGONM;
		else if  (!memicmp(argv[1],"DJ",3))  set_type = DRAGONJ;
		else  {
			printf("Invalid first parameter.\n");
			exit(0);
		}
	}
	else printf("Invalid parameters. See readme.txt for usage details.\n");

	if (argc > 2) infilename = argv[2];
	if (argc > 3) strcpy(outfilename,argv[3]);
	if (argc > 4) iterno = atoi(argv[4]);
	int_type = ((argc > 5) && (!memcmp(argv[5],"int",3)));


	/* specifical case:  initialization of starting set */
	if  (!memicmp(infilename,"_new",5)) {
		descr.screen_x =   max(10,atoi(argv[4]));
		descr.screen_y =   max(10,atoi(argv[5]));
		iterno = 128;
		//int_type = TRUE;

		if (graphinit() < 0) {printf("error graphics initializing."); exit(0);}
		switchmode(VGA256);
		switchmode(SVGA640x480x256);
		setDACblock(0,256,palette);

		switch (set_type) {
		case MANDELBROT:
			ms = new mandelbrot_set(iterno, int_type,
				descr.screen_x, descr.screen_y,
				-1.5,1.5,
				-1.5 *descr.screen_y / descr.screen_x,
				 1.5 *descr.screen_y / descr.screen_x);
			ms->set_coloring(0,5, NONE, 1,254,  LINEAR);
			julia_like = FALSE;
			break;

		case DRAGONM:
			ms = new dragon_m_set(iterno, int_type,
				descr.screen_x, descr.screen_y,
				-4.0,4.0,
				-4.0 *descr.screen_y / descr.screen_x,
				 4.0 *descr.screen_y / descr.screen_x);
			ms->set_coloring(1,1000, LINEAR, 0,0,  NONE);
			julia_like = FALSE;
			break;
		default: exit(0);
		}
		init_keeper(ms,pal);
		ms -> calculate_set();
		save_file (ms,&pal,outfilename);
		goto finish;
	}
	/* end of specifical case */



	if (graphinit() < 0) {printf("error graphics initializing."); exit(0);}
	switchmode(VGA256);
	switchmode(SVGA640x480x256);
	setDACblock(0,256,palette);

	grf_r = open_image_file(infilename);

	if (grf_r == NULL) {closegraph; exit(0);}
	ok = read_descr(infilename, &descr);
	if (!ok) { printf("Not a ...");
		closegraph;
		goto finish;
	}

	initmouse_easy();
	setmouserange(0,0,getmaxx(),getmaxy());

	restrictions.b.x = getmaxx();
	restrictions.b.y = getmaxy();

	switch (set_type) {
		case MANDELBROT:
			if (descr.set_type != set_type) type_error();
			ms = new mandelbrot_set(iterno, int_type,
				descr.screen_x, descr.screen_y,
				descr.x_min,descr.x_max, descr.y_min,descr.y_max);
			ms->set_coloring(0,5, NONE, 1,254,  LINEAR);
			julia_like = FALSE;
			break;

		case DRAGONM:
			if (descr.set_type != set_type) type_error();
			ms = new dragon_m_set(iterno, int_type,
				descr.screen_x, descr.screen_y,
				descr.x_min,descr.x_max, descr.y_min,descr.y_max);
			ms->set_coloring(1,1000, LINEAR, 0,0,  NONE);
			julia_like = FALSE;
			break;


		case JULIA:
			if (descr.set_type == set_type) no_mand_map = TRUE;
			else if (descr.set_type != MANDELBROT) type_error();
			ms = new julia_set(iterno,int_type,
				descr.screen_x, descr.screen_y,
				c_re, c_im,
				descr.x_min,descr.x_max, descr.y_min,descr.y_max);
			     ms->set_coloring(0,954, LINEAR, 0,0,  NONE);
			     julia_like = TRUE;
			break;

		case DRAGONJ:
			if (descr.set_type == set_type) no_mand_map = TRUE;
			else if (descr.set_type != DRAGONM) type_error();
			ms = new dragon_j_set(iterno,int_type,
				 descr.screen_x, descr.screen_y,
				 c_re, c_im,
				 descr.x_min,descr.x_max, descr.y_min,descr.y_max);
				ms->set_coloring(0,0, NONE, 0,255,  LINEAR);
			     julia_like = TRUE;
			break;
		}

	if (julia_like) {
		if (no_mand_map) {
			show_file(grf_r, ms->screen_left, ms->screen_top);
			ms->set_c(descr.c_re,descr.c_im);
		}
		else {	ok = mand_point(p,restrictions);
			if (!ok) goto finish;
			init_keeper(ms,pal);
			ms -> calculate_set();
	       }
	}
	else show_file(grf_r,ms->screen_left, ms->screen_top);

	fnsplit(outfilename,drive,dir,file,ext);
	outfilename[strlen(drive) + strlen(dir)] = 0;
	file[6] = 0;
	strcat(outfilename,file);
	strcat(outfilename,"00.tga");

	repeater = 0;
	do {
		  do  {
			if (repeater <= 0) key = clip_area(win, restrictions);
			else { if (repeater & 1) key = kwAltS;
			       else key = kwEnter;

				repeater--;
			}

			clipping = TRUE;  ok = TRUE;
			switch (key) {

			    case kwEsc:
			    case kwAltX:    ok = FALSE;
			    case kwEnter:  clipping = FALSE; break;
			    case kwAltS:   outfilename[strlen(outfilename) - 6] =
						'0' + saveindex / 10;
					    outfilename[strlen(outfilename) - 5] =
						'0' + saveindex % 10;
					save_file (ms,&pal,outfilename);
					saveindex++;
					break;
			    case kwBkSp: if ((!julia_like) || no_mand_map) break;
					 ok = mand_point(p,restrictions);
					 if (ok) {
						init_keeper(ms,pal);
						ms -> calculate_set();
					}
					break;
			    case kwF2:  repeater = 4; break;
			    case kwF3:  repeater = 6; break;
			    case kwF4:  repeater = 8; break;
			    case kwF5:  repeater = 10; break;
			    case kwF6:  repeater = 12; break;
			    }
		  } while (clipping);

		if (ok) {
			ms -> set_window(win);
			init_keeper(ms,pal);
			ms -> calculate_set();
		}
	} while(ok);


	delay_key1(1000);
	delete grf_r;
	delete(ms);
finish:	closegraph();
}
