/************************************************************************/
/* This file contains procedures for Mandelbrot-like set exploring      */
/*									*/
/*    E.Podvoysky from ^Z for WROX press book, 1994			*/
/************************************************************************/
#include <math.h>
#include <conio.h>
#include <string.h>

#include "common.h"
#include "graph.h"
#include "mand_set.h"

extern "C" int calc_one_point(double x,double y,int max_it,
			double &im_sq,double &re_sq);
extern "C" int calc_one_point_jul(double x,double y,int max_it,
			double &im_sq,double &re_sq,double c_image,
			double c_real);


mandelbrot_set::mandelbrot_set(int max_it_, BOOL int_calc_,
		int screen_x_, int screen_y_,
		float x_min_, float x_max_, float y_min_, float y_max_) {
	max_it = max_it_;
	int_calc = int_calc_;
	screen_x = screen_x_;
	screen_y = screen_y_;
	set_ranges(x_min_, x_max_, y_min_,  y_max_);
	row = new SET_POINT[screen_x];
	showrow = new BYTE[screen_x];
	order_color_map = NULL;
	set_coloring(0,0, NONE, 1,255, DETAILS);
	keeper = NULL;
	screen_left = (getmaxx() - screen_x)/2;
	screen_top  = (getmaxy() - screen_y)/2;
	calculated = FALSE;
}

mandelbrot_set::~mandelbrot_set() {
	delete row;
	if (order_color_map != NULL) delete order_color_map;
	delete showrow;
	if (keeper != NULL) delete keeper;
}

void mandelbrot_set::calc_one_row(double y){
	register int iter_no;
	int ix;
	double im, re, im_sq, re_sq, x, deltax;
	SET_POINT *rw;

	rw = row;
	x = x_min;
	deltax = dx; // really is faster then this -> x

	for (ix = screen_x; ix > 0 ; ix--,x += deltax) {
		iter_no = calc_one_point(x,y,max_it,im_sq,re_sq);

	   /*iter_no = max_it;
	   im = re = im_sq = re_sq = 0.0;
	   while ((im_sq + re_sq < 4.0) && (iter_no-- > 0)) {
		im =  2.0*re*im + y;
		re =  re_sq - im_sq + x;
		re_sq = re*re;
		im_sq = im*im;
	   }*/

	   (*rw).iterations = (iter_no+1);
	   (*rw).mod = im_sq + re_sq;
	   rw++;
	}
}

void mandelbrot_set::calc_one_row_int(double y) {
	register int iter_no;
	register ix;
	double x, deltax;
	register long im, re;
	long im_sq, re_sq, xl, yl;
	register SET_POINT *rw = row;

	// re,im are scaled with 2 power 13 = 0x2000
	// yl,xl are scaled with 2 power 26 = 0x4000000
	// abs(re),abs(im) < 2

	yl = y*0x4000000 +0x1000;
	x = x_min;
	deltax = dx;

	for (ix = 0; ix < screen_x; ix++,x += deltax) {
		//iter_no = calc_one_point(x,Y,max_it);

	   xl = x*0x4000000 +0x1000;

	   iter_no = max_it;
	   im = re = im_sq = re_sq = 0;
	   while ((im_sq + re_sq < 0x10000000) && (iter_no-- )) {
		im =  ( (re*im << 1) + yl) >> 13; // (/ 0x2000);
		re =  (re_sq - im_sq + xl) >> 13; // (/ 0x2000);
		re_sq = re*re;
		im_sq = im*im;
	   }

	   (*rw).iterations = max_it - (iter_no+1);
	   (*rw).mod = (float)(im_sq + re_sq)/ 0x4000000;
	   rw++;
	}
}


void mandelbrot_set::set_coloring(int chaos_col1_,int chaos_col2_,int chaos_scheme_,
		  int order_col1_,int order_col2_,int order_scheme_){
	int i, d;
	float f;

	chaos_col1 = chaos_col1_;
	chaos_col2 = chaos_col2_;
	chaos_scheme = chaos_scheme_;

	order_col1 =  order_col1_;
	order_col2 = order_col2_;
	order_scheme = order_scheme_;

	if (order_color_map == NULL) order_color_map = new BYTE[max_it];


	switch(order_scheme) {
	case DETAILS:
		d = order_col2-order_col1+1;
		for (i = 0; i < max_it; i++)
			order_color_map[i] = order_col1 + i % d;
		break;

	case LINEAR:
		f = (max_it-1.0) / (order_col2-order_col1);
		for (i = 0; i < max_it; i++)
			order_color_map[i] = order_col1 + i / f;
		break;

	case LOG:
		 f = log(max_it-1)/ (order_col2-order_col1);
		 for (i = 0; i < max_it; i++)
			order_color_map[i] = order_col1 + log(i)/f;
		 break;
	case NONE:  for (i = 0; i < max_it; i++) order_color_map[i] = order_col1;
	}

	switch(chaos_scheme) {
	case DETAILS:
		d = chaos_col2-chaos_col1+1;
		for (i = 0; i < 256; i++)
			chaos_color_map[i] = chaos_col1 + i % d;
		break;

	case LINEAR:
		f = 255.0 / (chaos_col2-chaos_col1);
		for (i = 0; i < 256; i++)
			chaos_color_map[i] = chaos_col1 + i / f;
		break;

	case LOG:
		 f = log(255)/ (chaos_col2-chaos_col1);
		 for (i = 0; i < 256; i++)
			chaos_color_map[i] = chaos_col1 + log(i+1)/f;
		 break;
	case NONE:  for (i = 0; i < 256; i++) chaos_color_map[i] = chaos_col1;
	}

}

void mandelbrot_set::calculate_set() {
	int iy;
	double dy,y;
	calculated = FALSE;

	dy = (y_max - y_min) / screen_y;
	dx = (x_max - x_min) / screen_x;
	for (iy = 0, y = y_min; iy < screen_y; iy++, y += dy) {
		if (kbhit()) return;
		if (int_calc)  calc_one_row_int(y);
		else calc_one_row(y);
		show_row(iy);
	}
	calculated = TRUE;
}

SET_TYPE mandelbrot_set::my_type() {return MANDELBROT;}

void mandelbrot_set::show_row(int iy) {
	int ix,col;
	register BYTE *sr;
	register SET_POINT *rw;

	rw = row;
	sr = showrow;
	for (ix = screen_x; ix > 0;   ix--) {
		if ((*rw).iterations < max_it)
			*(sr++) = order_color_map[(*(rw++)).iterations];
		else 	*(sr++) = chaos_color_map[(*(rw++)).mod * 64.0];
		}
	if (keeper != NULL) keeper->store_line(showrow);
	putrow(rowptr(showrow),screen_left,screen_top+iy,screen_x);
}

void mandelbrot_set::set_keeper(image_keeper *keeper_) {
if (keeper != NULL) delete keeper;
keeper = keeper_;
}

void mandelbrot_set::set_window(RECT win) {
	float coef;
	float x_min_, x_max_, y_min_, y_max_;

	coef =  (x_max - x_min)/screen_x;
	x_min_ =  x_min + (win.a.x - screen_left + 0.5) * coef;
	x_max_ =  x_min + (win.b.x - screen_left + 0.5) * coef;

	coef =  (y_max - y_min)/screen_y;
	y_min_ =  y_min + (win.a.y - screen_top + 0.5) * coef;
	y_max_ =  y_min + (win.b.y - screen_top + 0.5) * coef;
	set_ranges(x_min_, x_max_, y_min_, y_max_);
}
void mandelbrot_set::set_ranges(float x_min_, float x_max_,
		float y_min_, float y_max_) {
	x_min = x_min_;
	x_max = x_max_;
	y_min = y_min_;
	y_max = y_max_;

}

/////////////////////////////////////////////////////////////////////////
julia_set::julia_set(int max_it_, BOOL int_calc_,
		int screen_x_, int screen_y_,
		float c_re, float c_im,
		float x_min_, float x_max_, float y_min_, float y_max_):
	 mandelbrot_set(max_it_, int_calc_, screen_x_,  screen_y_,
			x_min_, x_max_,  y_min_,  y_max_)
{
set_c(c_re, c_im);
}


SET_TYPE julia_set::my_type() {return JULIA;}

void julia_set::calc_one_row(double y){
	int iter_no,ix;
	double im, re, im_sq, re_sq, x,deltax;
	register SET_POINT *rw = row;

	x = x_min;
	deltax = dx;
	for (ix = 0; ix < screen_x; ix++,x += deltax) {

		//im = y; re = x; im_sq = im*im; re_sq = re*re;
		iter_no = calc_one_point_jul(x,y,max_it,
						im_sq,re_sq,c_image,c_real);
	   /*iter_no = max_it;
	   while ((im_sq + re_sq < 4.0) && (iter_no-- )) {
		im =  2*re*im + c_image;
		re =  re_sq - im_sq + c_real;
		re_sq = re*re;
		im_sq = im*im;
	   }*/

	   (*rw).iterations = (iter_no+1);
	   (*rw).mod = im_sq + re_sq;
	   rw++;
	}
}

void julia_set::calc_one_row_int(double y) {
	int iter_no,ix;
	double x, deltax = dx;
	long im, re, im_sq, re_sq, xl, yl;
	register SET_POINT *rw = row;

	yl = y*0x2000+0.5;
	x = x_min;

	for (ix = 0; ix < screen_x; ix++,x += deltax ) {

	   xl = x*0x2000+0.5;

	   iter_no = max_it;
	   im = yl; re = xl;
	   re_sq = re*re; im_sq = im*im;

	   while ((im_sq + re_sq < 0x10000000) && (iter_no-- )) {
		im =  (2*re*im + c_image_long) >> 13; /// 0x2000;
		re =  (re_sq - im_sq + c_real_long) >> 13; /// 0x2000;
		re_sq = re*re;
		im_sq = im*im;

	   }
	   (*rw).iterations = max_it - (iter_no+1);
	   (*rw).mod = (float)(im_sq + re_sq)/ 0x4000000;
	   rw++;
	}
}

void julia_set::set_c(float c_re, float c_im) {
	c_real = c_re;
	c_image = c_im;
	c_real_long = c_real*0x4000000 + 0x1000;
	c_image_long = c_image*0x4000000 + 0x1000;
}

/////////////////////////////////////////////////////////////////////////
dragon_m_set::dragon_m_set(int max_it_, BOOL int_calc_,
		int screen_x_, int screen_y_,
		float x_min_, float x_max_, float y_min_, float y_max_):
	 mandelbrot_set(max_it_, int_calc_, screen_x_, screen_y_,
		x_min_, x_max_, y_min_, y_max_) {}

SET_TYPE dragon_m_set::my_type() {return DRAGONM;}

void dragon_m_set::calc_one_row(double y){
	register int iter_no;
	int ix;
	double im, re, im_sq, re_sq, x;
	double re2, im2;
	double mand_c_i, mand_c_r;
	SET_POINT *rw;

	rw = row;
	x = x_min;

	for (ix = screen_x; ix > 0 ; ix--, x += dx) {

		//iter_no = max_it;
		//im = re = im_sq = re_sq = 0.0; // assumed z0 = 1/2
	       // recalc from dragon to Mandelbrot parameters
		mand_c_r = 0.25*(2*x  - (x*x - y*y)); // c1 = (2-c)*c/4
		mand_c_i = 0.5*y*(1-x);

		iter_no = calc_one_point(mand_c_r,mand_c_i,max_it,im_sq,re_sq);
	   /*
	   while ((im_sq + re_sq < 4.0) && (iter_no-- )) {
		im =  2*re*im + mand_c_i;
		re =  re_sq - im_sq + mand_c_r;
		re_sq = re*re;
		im_sq = im*im;
	   }*/

	   (*rw).iterations = (iter_no+1);
	   (*rw).mod = im_sq + re_sq;
	   rw++;
	}
}

void dragon_m_set::calc_one_row_int(double y) {
dragon_m_set::calc_one_row(y);
};


dragon_j_set::dragon_j_set(int max_it_, BOOL int_calc_,
		int screen_x_, int screen_y_,
		float c_re, float c_im,
		float x_min_, float x_max_, float y_min_, float y_max_):

	 julia_set(max_it_, int_calc_, screen_x_, screen_y_,
		c_re, c_im,
		x_min_, x_max_, y_min_, y_max_) {}

SET_TYPE dragon_j_set::my_type() {return DRAGONJ;}

void dragon_j_set::calc_one_row(double y){
	int iter_no,ix;
	double im, re, im_sq, re_sq, x;
	double mand_c_i, mand_c_r;
	double c_re = c_real, c_im = c_image, z;
	register SET_POINT *rw = row;

	x = x_min;
		// recalc constant
	mand_c_r = 0.25*(2*c_re  - (c_re*c_re - c_im*c_im)); // c1 = (2-c)*c/4
	mand_c_i = 0.5*c_im*(1-c_re);

	for (ix = 0; ix < screen_x; ix++,x += dx) {
		//iter_no = calc_one_point(x,Y,max_it);

	   //iter_no = max_it;
	   // recalc  starting value z0 = -c*(z0 - 0.5)
	   z = x - 0.5; // real part
	   re = c_im * y - c_re * z;
	   im = -(c_im * z + y * c_re);
	   //im_sq = im*im; re_sq = re*re;

	   iter_no = calc_one_point_jul(re,im,max_it,
						im_sq,re_sq,mand_c_i,mand_c_r);

	   /*while ((im_sq + re_sq < 4.0) && (iter_no-- )) {
		im =  2*re*im + mand_c_i;
		re =  re_sq - im_sq + mand_c_r;
		re_sq = re*re;
		im_sq = im*im;
	   } */

	   (*rw).iterations = (iter_no+1);
	   (*rw).mod = im_sq + re_sq;
	   rw++;
	}
}

void dragon_j_set::calc_one_row_int(double y) {
dragon_j_set::calc_one_row(y);
};


void drive_color(colortype *pal, int  i1, int i2) {
	int i,dr,dg,db;
	colortype c1,c2;
	c1 = pal[i1];
	c2 = pal[i2];
	dr = c2.r - c1.r;
	dg = c2.g - c1.g;
	db = c2.b - c1.b;
	for (i = 1; i < i2 - i1; i++) {
	    pal[i1+i].r = c1.r + (dr*i)/(i2-i1);
	    pal[i1+i].g = c1.g + (dg*i)/(i2-i1);
	    pal[i1+i].b = c1.b + (db*i)/(i2-i1);
	}
}

void init_palette(colortype *pal, int colors, colortype *colordef) {
	int j1,j2,i;
	pal[0] =  colordef[0];
	j1 = 0;
	for (i = 1; i < colors; i++) {
		j2 = (i*254)/(colors-1);
		pal[j2] = colordef[i];
		drive_color(pal,j1,j2);
		j1 = j2;
	}
}
